import datetime
import os
import random
import sys
import time
from abc import abstractmethod

import cv2
import matplotlib
import numpy as np
import onnxruntime
import torch
from torch.backends import cudnn

from ftools.GLOBAL_LOG import flog
from ftools.ai.fits.f_fit_tools import SmoothedValue, FitExecutorBase, FitExecutor4OD, FitExecutor4KPS
from ftools.ai.picture.f_show import f_show_od_np4plt_v3, draw_text_chinese4cv
from ftools.f_general import get_img_file
from ftools.ai.fits.f_predictfun import batch_nms_v2
from ftools.general.fsys.fcamera import init_video


class DispatchTask:
    def __init__(self, fun_task_cfg=None) -> None:
        super().__init__()
        self.init_sys(None)
        self.fun_task_cfg = fun_task_cfg

    def init_sys(self, cfg):
        # -----------通用系统配置----------------
        # format short g, %precision=5
        torch.set_printoptions(linewidth=320, sci_mode=False, precision=5, profile='long')
        np.set_printoptions(linewidth=320, suppress=True, formatter={'float_kind': '{:11.5g}'.format})
        # 修改配置文件
        matplotlib.rcParams['font.sans-serif'] = ['SimHei']  # 设置matplotlib可以显示汉语
        matplotlib.rcParams['axes.unicode_minus'] = False
        matplotlib.rcParams['font.size'] = 11
        # matplotlib.rc('font', **{'size': 11})

        # try:
        #     # 对windwos进行限制防止出错  by yolox
        #     import resource  # 这个模块没有
        #     rlimit = resource.getrlimit(resource.RLIMIT_NOFILE)
        #     resource.setrlimit(resource.RLIMIT_NOFILE, (8192, rlimit[1]))
        # except Exception as e:
        #     # Exception might be raised in Windows OS or rlimit reaches max limit number.
        #     # However, set rlimit value might not be necessary.
        #     raise e

        # --- pytorch 系统加速 ---
        torch.multiprocessing.set_sharing_strategy('file_system')  # 多GPU必须开 实测加速
        os.environ["OPENCV_OPENCL_RUNTIME"] = "disabled"  # by yolox
        try:
            # by yolox
            cv2.setNumThreads(0)  # 加速 防止OpenCV进入多线程(使用PyTorch DataLoader)
            cv2.ocl.setUseOpenCL(False)
        except Exception as e:
            # cv2 version mismatch might rasie exceptions.
            raise e
        seed = 0
        # 随机种子
        np.random.seed(seed)
        random.seed(seed)
        # from yolo5 Speed-reproducibility tradeoff https://pytorch.org/docs/stable/notes/randomness.html
        torch.manual_seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)
            cudnn.deterministic = True
            if seed == 0:  # slower, more reproducible
                cudnn.benchmark, cudnn.deterministic = False, True
            else:  # faster, less reproducible
                cudnn.benchmark, cudnn.deterministic = True, False

    def show_args(self, cfg):
        '''
        重要参数显示
        本方法 参数在 cfg 中具备较强依赖  不适合通用

        :param cfg:
        :return:
        '''
        if cfg.IS_WRITER:
            if cfg.KEEP_WRITER is not None:
                tb_writer_info = cfg.KEEP_WRITER
            else:
                tb_writer_info = '开,随机目录'
        else:
            tb_writer_info = False

        if cfg.IS_WARMUP:
            warmup_info = cfg.NUM_WARMUP
        else:
            warmup_info = False

        if cfg.IS_MULTI_SCALE:
            size = '多尺度: ' + str(cfg.MULTI_SCALE_VAL)
        else:
            size = cfg.SIZE_WH_INPUT_TRAIN

        _text_dict = {
            'dataset_name': cfg.DATA_INFO.dataset_name,
            'tb_writer_info': tb_writer_info,
            'device': cfg.DEVICE,
            'num_vis_z': cfg.NUM_VIS_Z,
            'num_workers': cfg.NUM_WORKERS,
            'path_save_weight': cfg.PATH_SAVE_WEIGHT,
            'warmup_info': warmup_info,
            'is_visual': cfg.IS_VISUAL,
            'size': size,
            'batch_train': cfg.DATA_INFO.batch_train,
            'mode_train': cfg.DATA_INFO.mode_train,
            'ema': cfg.IS_EMA,
            'maps_def_max': cfg.MAPS_DEF_MAX,
            'is_mosaic': cfg.IS_MOSAIC,
            'is_mixup': cfg.IS_MIXUP,
        }

        separator = '  \n'
        for k, v in _text_dict.items():
            _text_dict[k] = str(v) + separator

        # 遍历转str + 分隔

        # _text = '--- 训练类 ---\n'.format(**_text_dict)

        _text1 = '----------------- 系统类 ---------------\n' \
                 '\t 当前数据集: {dataset_name}' \
                 '\t 当前设备: {device}' \
                 '\t 权重路径: {path_save_weight}' \
                 '\t tb_writer: {tb_writer_info}' \
                 '\t num_workers: {num_workers}' \
                 '\t WARMUP: {warmup_info}' \
                 '\t IS_VISUAL: {is_visual}' \
                 '\t size: {size}' \
                 '\t 上限AP: {maps_def_max}'

        _text2 = '----------------- 训练类 -----------------\n' \
                 '\t batch:{batch_train}' \
                 '\t mode_train:{mode_train}' \
                 '\t ema:{ema}' \
                 '\t IS_MOSAIC:{is_mosaic}' \
                 '\t IS_MIXUP:{is_mixup}'

        _text3 = '----------------- 测试类 -----------------\n' \
                 '\t NUM_VIS_Z:{num_vis_z}'

        show_text = (_text1 + _text2 + _text3).format(**_text_dict)
        print(show_text)

    def run_train(self):
        t00 = time.time()
        cfg, model, fun_loss, fun_test, dataloader_train, dataloader_test \
            = self.fun_task_cfg()

        self.show_args(cfg)

        ''' --------------- 启动 --------------------- '''
        # 'segm', 'bbox', 'keypoints'
        if cfg.DATA_INFO.mode_test == 'keypoints':
            executor = FitExecutor4KPS(cfg, model=model,
                                       fun_loss=fun_loss, fun_test=fun_test,
                                       dataloader_train=dataloader_train,
                                       dataloader_test=dataloader_test, )
            pass
        elif cfg.DATA_INFO.mode_test == 'bbox':
            executor = FitExecutor4OD(cfg, model=model,
                                      fun_loss=fun_loss, fun_test=fun_test,
                                      dataloader_train=dataloader_train,
                                      dataloader_test=dataloader_test,
                                      )
        elif cfg.DATA_INFO.mode_test == 'segm':
            raise Exception('暂不支持 %s' % cfg.DATA_INFO.mode_test)
        else:
            raise Exception('暂不支持 %s' % cfg.DATA_INFO.mode_test)

        executor.frun()

        time_text = str(datetime.timedelta(seconds=int(time.time() - t00)))
        return time_text


''' ----------------- 检测任务执行者 --------------------- '''


class DetectTaskBase:

    def __init__(self, mode_detect, data_transform, device, ids_classes,
                 model=None, path_pic=None, video_address=0, is_show_res=True,
                 ) -> None:
        '''

        :param mode_detect: video  pic  webcam
        :param data_transform:
        :param device:
        :param ids_classes:
        :param path_pic:
        :param video_address: 视频地址
        '''
        super().__init__()
        self.model = model
        self.mode_detect = mode_detect
        self.data_transform = data_transform
        self.device = device
        self.ids_classes = ids_classes
        self.path_pic = path_pic
        self.video_address = video_address
        self.is_show_res = is_show_res

        self._init_args()

    def detect(self):
        if self.mode_detect == 'pic':
            files_pic = self.examine_path_pic()
            for file in files_pic:
                # 在 reses 中处理返回值
                # 计时在里面单独计时从打开文件开始 因为显示时间除外
                reses = self.detect4pic(file=file)

        elif self.mode_detect == 'video':
            cap = init_video(det=self.video_address)
            self.is_show_res = False  # 视频时图形显示 必须关闭
            reses = self.detect4video(cap=cap)

        flog.info('%s 模式 detect 运行完成' % self.mode_detect)

    def examine_path_pic(self):
        if self.path_pic is not None and os.path.exists(self.path_pic):
            if os.path.isdir(self.path_pic):
                files_pic = get_img_file(self.path_pic)
            elif os.path.isfile(self.path_pic):
                files_pic = [self.path_pic]
            else:
                raise Exception('path_pic 有问题 %s' % self.path_pic)
        else:
            raise Exception('path_pic 不存在 %s' % self.path_pic)

        if len(files_pic) == 0:
            raise Exception('path_pic 路径没有图片 %s' % self.path_pic)

        return files_pic

    def _init_args(self):
        matplotlib.rcParams['font.sans-serif'] = ['SimHei']  # 设置matplotlib可以显示汉语
        matplotlib.rcParams['axes.unicode_minus'] = False
        matplotlib.rcParams['font.size'] = 11

    def get_reses_z(self, img_np_file, reses_detect):
        '''
        强关联  get_detect_reses 方法 否则复写
        :param reses_detect:
        :return:
        '''
        if reses_detect is None:
            # flog.debug('没有目标 %s', )
            # sys.exit(-1)
            return None
        else:
            # 一次只有一张 ids_batch 无用
            ids_batch, p_ltrb_one, p_labels, p_scores = reses_detect
            if len(p_labels) == 0:
                # 没有目标
                return None

            size_wh_file = np.array(img_np_file.shape[:2][::-1])
            size_wh_file_x2 = np.tile(size_wh_file, 2)
            # if isinstance(p_ltrb_one, torch.Tensor):
            #     size_wh_file_x2 = torch.tensor(size_wh_file_x2)
            p_boxes_ltrb_f = p_ltrb_one * size_wh_file_x2
            # 这个待后续可处理成json返回
            title_text, p_texts = self.cre_title_text(size_wh_file=size_wh_file,
                                                      p_boxes_ltrb_f=p_boxes_ltrb_f,
                                                      p_labels=p_labels,
                                                      p_scores=p_scores,
                                                      img_np_file=img_np_file,
                                                      )

        return p_boxes_ltrb_f, p_labels, p_scores, title_text, p_texts

    def cre_title_text(self, size_wh_file, p_boxes_ltrb_f, p_labels, p_scores, img_np_file, ):
        '''
        这个用于创建检测结果的文本
        :return:
        '''
        title_text = '%s x %s (num_pos = %s) max=%s' % (str(size_wh_file[0]),  # w
                                                        str(size_wh_file[1]),  # h
                                                        str(len(p_scores)),
                                                        str(round(p_scores.max().cpu().item(), 2))
                                                        )
        p_texts = []
        for i, p_label in enumerate(p_labels):
            s = self.ids_classes[str(p_label.long().item())] + ':' + str(round(p_scores[i].cpu().item(), 2))
            p_texts.append(str(s))

        if self.is_show_res:
            f_show_od_np4plt_v3(img_np_file, p_ltrb=p_boxes_ltrb_f,
                                title=title_text, p_texts=p_texts,
                                is_recover_size=False)
        return title_text, p_texts

    def print_time(self, t00, p_labels):
        t01 = time.time()
        _text_dict = {
            'time': t01 - t00,
            'num_pos': 0 if p_labels is None else len(p_labels)
        }
        print('检测时间: {time:0.4f} 共有 {num_pos} 个目标'.format(**_text_dict))

    @abstractmethod
    def detect4pic(self, file):
        pass

    def detect4video(self, cap):
        fps = 0.0
        count = 0
        num_out = 0
        while cap.isOpened():
            start_time = time.time()
            '''---------------数据加载及处理--------------'''
            ref, img_np_file = cap.read()  # 读取某一帧 ref是否成功
            if not ref:
                if num_out >= 3:
                    raise Exception('cap 读取出错~~~num_out=%s' % num_out)
                num_out += 1
                flog.error('cap 读取出错，再试')
                continue

            reses_detect = self.get_detect_reses(img_np_file)
            reses_z = self.get_reses_z(img_np_file, reses_detect)
            if reses_z is None:
                img_np = img_np_file
                title_text = 'no target!!!'
            else:
                p_boxes_ltrb_f, p_labels, p_scores, title_text, p_texts = reses_z
                if len(p_labels) == 0:
                    img_np = img_np_file
                    title_text = 'no target!!!'
                else:
                    # cv转码 乱码
                    # if isinstance(p_boxes_ltrb_f, torch.Tensor):
                    # ts -> np
                    p_boxes_ltrb_f = p_boxes_ltrb_f.numpy()

                    # 这里是画框
                    img_np = img_np_file
                    for box, text in zip(p_boxes_ltrb_f, p_texts):
                        box = box.astype(np.int)
                        cv2.rectangle(img_np, (box[0], box[1]), (box[2], box[3]), (0, 0, 255), 2)  # (l,t),(r,b),颜色.宽度
                        left = box[0]
                        top = box[1]  # putText 这个是文字中间
                        img_np = draw_text_chinese4cv(img_np, text, left, top)
                        # cv2.putText(img_np, text, (cx, cy), cv2.FONT_HERSHEY_DUPLEX, 0.7, (255, 255, 255), )

            # 这个 img_np 就是rgb 不用转
            # img_np = cv2.cvtColor(img_np, cv2.COLOR_RGB2BGR)

            # print("fps= %.2f" % (fps))
            count += 1
            text_ = "%s fps= %.2f count=%s " % (title_text, fps, count)
            # 图片，添加的文字，左上角坐标，字体，字体大小，颜色，字体粗细
            img_np = cv2.putText(img_np, text_, (0, 20), cv2.FONT_HERSHEY_SIMPLEX, 0.5,
                                 (0, 255, 0), 1)
            # img_np=draw_text_chinese4cv(img_np, text_, 0, 40, )
            # 极小数
            fps = (fps + (1. / max(sys.float_info.min, time.time() - start_time))) / 2

            # cv2.imshow(title_text, img_np)
            # cv2.imshow('无效'.encode("gbk").decode(errors="ignore"), img_np) # 无效
            cv2.imshow(u'abc', img_np)

            c = cv2.waitKey(1) & 0xff  # 输入esc退出
            if c == 27:
                cap.release()
                cv2.destroyAllWindows()
                break

    @abstractmethod
    def get_detect_reses(self, img_np_file):
        '''
        必须返回以下内容
        :param img_np_file:
        :return:
            ids_batch, p_ltrbs, p_labels, p_scores = reses_detect
        '''
        pass


class DetectTaskFlowBase(DetectTaskBase):

    def __init__(self, mode_detect, data_transform, device, ids_classes,
                 path_pic=None,
                 pred_conf_thr=0.01,
                 pred_select_topk=500,
                 pred_nms_thr=0.65,
                 pred_nms_keep=9999,
                 is_show_res=True,
                 ) -> None:
        super(DetectTaskFlowBase, self).__init__(mode_detect=mode_detect,
                                                 data_transform=data_transform,
                                                 device=device,
                                                 ids_classes=ids_classes,
                                                 model=None,
                                                 path_pic=path_pic,
                                                 is_show_res=is_show_res,
                                                 )
        self.pred_conf_thr = pred_conf_thr
        self.pred_select_topk = pred_select_topk
        self.pred_nms_thr = pred_nms_thr
        self.pred_nms_keep = pred_nms_keep

    @abstractmethod
    def p_init(self, model_outs, targets):
        '''
        p_init 处理数据
        :param model_outs:
        :param targets: 这个需要处理  例如 to.显卡
        :return: 只需返回 model_outs
        '''
        return model_outs

    @abstractmethod
    def get_pscores(self, outs_init, targets):
        '''
        这个返回分数
        :param outs_init: p_init 处理后
        :param targets: 这个需要处理  例如 to.显卡
        :return:
            pscores : 这个用于判断 生成 mask_pos
            plabels : 这个用于二阶段
        '''
        pass

    @abstractmethod
    def get_stage_res(self, outs_init, imgs_ts_4d):
        '''

        :param outs_init:
        :return:  目标检测返回 这4个结果
            ids_batch2, p_boxes_ltrb2, p_labels2, p_scores2
        '''
        return None

    @abstractmethod
    def get_model_outs(self, imgs_ts_4d):
        '''
        模型输出的返回 这个由框架实现
        :param imgs_ts_4d:
        :return:
        '''
        pass

    def detect4pic(self, file):
        t00 = time.time()
        img_np_file = cv2.imread(file)  # h,w,c bgr

        reses_detect = self.get_detect_reses(img_np_file)
        reses_z = self.get_reses_z(img_np_file, reses_detect)
        if reses_z is None:
            return None

        p_boxes_ltrb_f, p_labels, p_scores, title_text, p_texts = reses_z
        self.print_time(t00, p_labels)
        return reses_z

    def get_detect_reses(self, img_np_file):
        '''

        :param img_np_file:
        :return:
            ids_batch, p_ltrbs, p_labels, p_scores = reses_detect
        '''
        # 这个必须返回一个target 这里返回的都是CPU值 框架会自动创建一个  这里
        img_ts_3d, target = self.data_transform(img_np_file)
        device = img_ts_3d.DEVICE

        # batch为1 一个图有一个 off_ltrb
        if target is not None and 'off_ltrb' in target:  # 多尺度归一
            off_ltrb_ts = torch.tensor(target['off_ltrb'], dtype=torch.float, device=device).unsqueeze(0)
        else:
            off_ltrb_ts = torch.zeros(1, 4, device=device, dtype=torch.float)

        # np只能用cpu()
        imgs_ts_4d = img_ts_3d.unsqueeze(0)
        model_outs = self.get_model_outs(imgs_ts_4d)

        # 转 tensor 保持一致
        reses_detect = self.postprocess(model_outs, imgs_ts_4d=imgs_ts_4d,
                                        targets=[target], off_ltrb_ts=off_ltrb_ts)

        return reses_detect

    def postprocess(self, model_outs, imgs_ts_4d, targets=None, off_ltrb_ts=None):
        outs = self.p_init(model_outs, targets)  # 1 -- 处理 np 转换
        pscores, plabels = self.get_pscores(outs, targets)  # 2 -- 处理 pscores, plabels
        mask_pos = pscores > self.pred_conf_thr
        if not torch.any(mask_pos):
            return None

        if pscores.shape[-1] > self.pred_select_topk:  # 并行取top100 与mask_pos 进行and操作
            # 每个批次取 topk 个
            ids_topk = pscores.topk(self.pred_select_topk, dim=-1)[1]  # torch.Size([32, 1000])
            mask_topk = torch.zeros_like(mask_pos)
            mask_topk[torch.arange(ids_topk.shape[0])[:, None], ids_topk] = True
            mask_pos = torch.logical_and(mask_pos, mask_topk)

        # 3 -- 处理 p_ltrb_input
        p_ltrb_input = self.get_stage_res(outs, imgs_ts_4d)

        size_hw_input = imgs_ts_4d.shape[-2:]
        device = imgs_ts_4d.DEVICE

        # off_ltrb_ts 实际归一化尺寸 偏移
        size_wh_input_ts = torch.tensor(list(size_hw_input)[::-1], device=device)
        # [2] ^^ [batch,4] ... [batch,2] ->  [batch,2] 实际可用图片尺寸
        size_wh_input_ts_x2 = (size_wh_input_ts - off_ltrb_ts[:, 2:] - off_ltrb_ts[:, :2]).repeat(1, 2)
        # [batch,dim,4] ^^ [batch,4] 平移选框只考虑lt           [batch,2] -> [batch,4] [batch,1,4]
        p_ltrb_one1 = (p_ltrb_input - off_ltrb_ts[:, :2].repeat(1, 2).unsqueeze(1)) / size_wh_input_ts_x2.unsqueeze(1)

        p_ltrb_one1 = p_ltrb_one1[mask_pos]
        p_ltrb_one1.clamp_(0, 1)

        ids_batch1, _ = torch.where(mask_pos)
        p_scores1 = pscores[mask_pos]
        p_labels1 = plabels[mask_pos]
        p_labels1 = p_labels1 + 1

        ids_batch2, p_ltrb_one2, p_labels2, p_scores2 = batch_nms_v2(
            ids_batch1=ids_batch1, p_ltrb1=p_ltrb_one1,
            p_labels1=p_labels1, p_scores1=p_scores1,
            threshold_nms=self.pred_nms_thr,
            num_max=self.pred_nms_keep,
        )
        return ids_batch2, p_ltrb_one2, p_labels2, p_scores2


class DetectTask_py(DetectTaskFlowBase):

    def __init__(self, mode_detect, data_transform, device, ids_classes, model,
                 path_pic=None,
                 pred_conf_thr=0.01,
                 pred_select_topk=500,
                 pred_nms_thr=0.65,
                 pred_nms_keep=9999,
                 is_show_res=True, ) -> None:
        super(DetectTask_py, self).__init__(mode_detect=mode_detect,
                                            data_transform=data_transform,
                                            device=device,
                                            ids_classes=ids_classes,
                                            path_pic=path_pic,
                                            pred_conf_thr=pred_conf_thr,
                                            pred_select_topk=pred_select_topk,
                                            pred_nms_thr=pred_nms_thr,
                                            pred_nms_keep=pred_nms_keep,
                                            is_show_res=is_show_res,
                                            )
        self.model = model

    def get_model_outs(self, imgs_ts_4d):
        self.model.eval()
        with torch.no_grad():
            model_outs = self.model(imgs_ts_4d)
        return model_outs


class DetectTaskONNX(DetectTaskFlowBase):

    def __init__(self, file_model, mode_detect, data_transform, device, ids_classes,
                 path_pic=None,
                 pred_conf_thr=0.01,
                 pred_select_topk=500,
                 pred_nms_thr=0.65,
                 pred_nms_keep=9999,
                 is_show_res=True,
                 ) -> None:
        '''
        暂时只支持CPU
        '''
        super(DetectTaskONNX, self).__init__(mode_detect=mode_detect,
                                             data_transform=data_transform,
                                             device=device,
                                             ids_classes=ids_classes,
                                             path_pic=path_pic,
                                             pred_conf_thr=pred_conf_thr,
                                             pred_select_topk=pred_select_topk,
                                             pred_nms_thr=pred_nms_thr,
                                             pred_nms_keep=pred_nms_keep,
                                             is_show_res=is_show_res,
                                             )
        self._init_model(file_model)
        self.file_model = file_model

    def _init_model(self, file_model):
        if not os.path.exists(file_model):
            raise Exception('file_model 不存在 {}'.format(file_model))

        onnx_session = onnxruntime.InferenceSession(file_model)
        self.output_names = [o.name for o in onnx_session.get_outputs()]
        self.input_names = [i.name for i in onnx_session.get_inputs()]
        self.model = onnx_session

    def get_model_outs(self, imgs_ts_4d):
        img_np_4d = imgs_ts_4d.numpy()
        model_outs = self.model.run(output_names=self.output_names,
                                    input_feed={self.input_names[0]: img_np_4d})
        return model_outs



from collections import OrderedDict

import torch
from torchvision import models
import torch.nn as nn

from ftools.GLOBAL_LOG import flog
from ftools.ai.fmodels.f_layer_get import ModelOuts4Resnet


class Net3(torch.nn.Module):
    def __init__(self):
        super(Net3, self).__init__()
        self.conv1 = torch.nn.Sequential()
        self.conv1.add_module("conv1", torch.nn.Conv2d(3, 32, 3, 1, 1))
        self.conv1.add_module("relu1", torch.nn.ReLU())
        self.conv1.add_module("pool1", torch.nn.MaxPool2d(2))

        self.conv2 = torch.nn.Conv2d(3, 32, 3, 1, 1)

        self.dense1 = torch.nn.Sequential()
        self.dense1.add_module("dense1", torch.nn.Linear(32 * 3 * 3, 128))
        self.dense1.add_module("relu2", torch.nn.ReLU())
        self.dense1.add_module("dense2", torch.nn.Linear(128, 10))

        self.dense2 = torch.nn.Sequential(
            torch.nn.Linear(32 * 3 * 3, 128),
            torch.nn.ReLU(),
            torch.nn.Linear(128, 10)
        )

        self.conv3 = torch.nn.Sequential(
            OrderedDict(
                [
                    ("conv1", torch.nn.Conv2d(3, 32, 3, 1, 1)),
                    ("relu1", torch.nn.ReLU()),
                    ("pool1", torch.nn.MaxPool2d(2))
                ]
            ))

    def forward(self, x):
        conv_out = self.conv1(x)
        res = conv_out.view(conv_out.size(0), -1)
        out = self.dense(res)
        return out


if __name__ == '__main__':
    model = models.resnet50(pretrained=True)
    dims_out = (512, 1024, 2048)
    model = ModelOuts4Resnet(model, dims_out)

    # for i, (name, param) in enumerate(model.named_parameters()):
    #     # 深度优化钻取
    #     print('---------- %s ----------' % i,
    #           'name =', name,
    #           '      ', param.shape)

    for i, (name, child) in enumerate(model.named_children()):
        # 这个不能钻取  获取的是下一级的 nn.Module  例如name <class 'torchvision.models.resnet.ResNet'>
        print('---------- %s ----------' % i)
        flog.debug('name %s', name)
        print(child)

    # for i, c in enumerate(model.children()):
    #     # 这个不能钻取
    #     print('---------- %s ----------' % i)
    #     print(c)

    # for i, m in enumerate(model.modules()):
    for i, (name, m) in enumerate(model.named_modules()):
        # 深度优先 这个是钻取 nn.Module <class 'f_pytorch.tools_model.f_layer_get.ModelOuts4Resnet'>
        print('---------- %s ----------' % i)
        pass

    for i, p in enumerate(model.parameters()):
        # 遍历出每一个参数 p这个是自定钻取
        print('---------- %s ----------' % i)
        pass



import os

import torch
import torch.nn.functional as F
import numpy as np

from ftools.ai.datas.dataset_coco import CustomCocoDataset, HeightenCocoDataset


def fcre_dataload(is_multi_scale, num_workers,
                  mode, batch, file_json, path_img, transform, name,
                  is_mosaic=False, is_mixup=False, is_mosaic_filter=False, thr_wh=5, thr_area=25,
                  size_hw_img=None, is_jump=False,  # is_mosaic 时必须传
                  kps_seq=None, is_debug=False,
                  ):
    if is_mosaic or is_mixup:
        dataset = HeightenCocoDataset(
            file_json=file_json,
            path_img=path_img,
            mode=mode,
            transform=transform,
            mode_balance_data=None,
            name=name,
            is_mosaic=is_mosaic,
            is_mixup=is_mixup,
            is_mosaic_filter=is_mosaic_filter,
            thr_wh=thr_wh,
            thr_area=thr_area,
            size_hw_img=size_hw_img,
            kps_seq=kps_seq,
            is_debug=is_debug,
        )
    else:
        dataset = CustomCocoDataset(
            file_json=file_json,
            path_img=path_img,
            mode=mode,
            transform=transform,
            mode_balance_data=None,
            name=name,
            is_jump=is_jump,
            kps_seq=kps_seq,
            is_debug=is_debug,
        )
    dataloader = torch.utils.data.DataLoader(
        dataset,
        batch_size=batch,
        num_workers=num_workers,
        shuffle=True,
        pin_memory=True,  # 不使用虚拟内存 GPU要报错
        # drop_last=True,  # 除于batch_size余下的数据
        # collate_fn=lambda x: fun_dataloader(x, cfg),
        collate_fn=CLS4collate_fn(is_multi_scale),
    )
    return dataloader


class CLS4collate_fn:
    # fdatas_l1
    def __init__(self, is_multi_scale):
        self.is_multi_scale = is_multi_scale

    def __call__(self, batch_datas):
        '''
        在这里重写collate_fn函数
        batch_datas: tuple[[tensor_img,dict_targets],...,[tensor_img,dict_targets]]
        '''
        # 训练才进这里
        if self.is_multi_scale:
            batch = len(batch_datas)
            imgs_list = []
            targets_list = []
            # 打开 tuple 数据
            for i, (img_ts, target) in enumerate(batch_datas):
                # flog.warning('fun4dataloader测试  %s %s %s ', target, len(target['boxes']), len(target['labels']))
                imgs_list.append(img_ts)
                targets_list.append(target)

            pad_imgs_list = []

            # 这里的最大一定能被32整除
            h_list = [int(s.shape[1]) for s in imgs_list]
            w_list = [int(s.shape[2]) for s in imgs_list]
            max_h = np.array(h_list).max()
            max_w = np.array(w_list).max()
            # self.cfg.tcfg_batch_size = [max_w, max_h] # 这样用多进程要报错
            for i in range(batch):
                img_ts = imgs_list[i]
                # 右下角添加 target 无需处理
                img_ts_pad = F.pad(img_ts, (0, int(max_w - img_ts.shape[2]), 0, int(max_h - img_ts.shape[1])), value=0.)
                pad_imgs_list.append(img_ts_pad)

                # debug 代码
                # fshow_pic_ts4plt(pad_img)  # 可视化 前面不要归一化
                # fshow_kp_ts4plt(pad_img,
                #                 targets_list[i]['boxes'],
                #                 targets_list[i]['kps'],
                #                 mask_kps=targets_list[i]['kps_mask'],
                #                 is_recover_size=False
                #                 )  # 可视化
                # f_show_od_ts4plt(img_ts_pad, targets_list[i]['boxes'], is_recover_size=False)
                # print('多尺度%s ' % str(pad_img.shape))

            imgs_ts_4d = torch.stack(pad_imgs_list)
        else:
            imgs_ts_3d = batch_datas[0][0]
            # 包装整个图片数据集 (batch,3,416,416) 转换到显卡
            imgs_ts_4d = torch.empty((len(batch_datas), *imgs_ts_3d.shape)).to(imgs_ts_3d)
            targets_list = []
            for i, (img, target) in enumerate(batch_datas):
                # flog.warning('fun4dataloader测试  %s %s %s ', target, len(target['boxes']), len(target['labels']))
                imgs_ts_4d[i] = img
                targets_list.append(target)
        return imgs_ts_4d, targets_list


def t001(cfg):
    from ftools.ai.datas.dta_heighten.f_data_pretreatment4np import ftransform_more_train, ftransform_nanodet_train
    path_host = ''
    mode = 'bbox'
    path_root = os.path.join(path_host, '/AI/datas/VOC2007')
    # path_img = os.path.join(path_root, 'train/JPEGImages')
    # file_json = os.path.join(path_root, 'coco/annotations/instances_type3_train_1066.json')
    # path_img = os.path.join(path_root, 'val/JPEGImages')
    # file_json = os.path.join(path_root, 'coco/annotations/instances_type3_val_413.json')
    path_img = os.path.join(path_root, 'train/JPEGImages')
    file_json = os.path.join(path_root, 'coco/annotations/instances_type3_train_1066.json')
    _transform_train = ftransform_more_train(cfg.SIZE_WH_INPUT_TRAIN)
    _transform_test = ftransform_nanodet_train(cfg.SIZE_WH_INPUT_TRAIN)
    cfg.NUM_CLASSES = 3
    dataset = CustomCocoDataset(
        file_json=file_json,
        path_img=path_img,
        mode=mode,
        transform=_transform_test,
        mode_balance_data=None,
    )
    dataloader_train = torch.utils.data.DataLoader(
        dataset,
        batch_size=32,
        num_workers=0,
        shuffle=True,
        pin_memory=True,  # 不使用虚拟内存 GPU要报错
        # drop_last=True,  # 除于batch_size余下的数据
        # collate_fn=lambda x: fun_dataloader(x, cfg),
        collate_fn=CLS4collate_fn(False),
    )
    return dataloader_train



import torch
from torch import nn


class SiLU(nn.Module):
    """export-friendly version of nn.SiLU()"""

    # @staticmethod
    def forward(self,x):
        return x * torch.sigmoid(x)


def get_activation(name="silu", inplace=True):
    if name == "silu":
        module = SiLU()
    elif name == "relu":
        module = nn.ReLU(inplace=inplace)
    elif name == "lrelu":
        module = nn.LeakyReLU(0.1, inplace=inplace)
    else:
        raise AttributeError("Unsupported act type: {}".format(name))
    return module


class BaseConv(nn.Module):
    """A Conv2d -> Batchnorm -> silu/leaky relu block"""

    def __init__(self, in_channels, out_channels, ksize, stride, groups=1, bias=False, act="silu"):
        super().__init__()
        # same padding
        pad = (ksize - 1) // 2
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=ksize,
            stride=stride,
            padding=pad,
            groups=groups,
            bias=bias,
        )
        self.bn = nn.BatchNorm2d(out_channels)
        self.act = get_activation(act, inplace=True)

    def forward(self, x):
        return self.act(self.bn(self.conv(x)))

    def fuseforward(self, x):
        return self.act(self.conv(x))


class DWConv(nn.Module):
    """Depthwise Conv + Conv"""

    def __init__(self, in_channels, out_channels, ksize, stride=1, act="silu"):
        super().__init__()
        self.dconv = BaseConv(
            in_channels, in_channels, ksize=ksize,
            stride=stride, groups=in_channels, act=act
        )
        self.pconv = BaseConv(
            in_channels, out_channels, ksize=1,
            stride=1, groups=1, act=act
        )

    def forward(self, x):
        x = self.dconv(x)
        return self.pconv(x)


class Bottleneck(nn.Module):
    # Standard bottleneck
    def __init__(
            self, in_channels, out_channels, shortcut=True,
            expansion=0.5, depthwise=False, act="silu"
    ):
        super().__init__()
        hidden_channels = int(out_channels * expansion)
        Conv = DWConv if depthwise else BaseConv
        self.conv1 = BaseConv(in_channels, hidden_channels, 1, stride=1, act=act)
        self.conv2 = Conv(hidden_channels, out_channels, 3, stride=1, act=act)
        self.use_add = shortcut and in_channels == out_channels

    def forward(self, x):
        y = self.conv2(self.conv1(x))
        if self.use_add:
            y = y + x
        return y


class SPPBottleneck(nn.Module):
    """Spatial pyramid pooling layer used in YOLOv3-SPP"""

    def __init__(self, in_channels, out_channels, kernel_sizes=(5, 9, 13), activation="silu"):
        super().__init__()
        hidden_channels = in_channels // 2
        self.conv1 = BaseConv(in_channels, hidden_channels, 1, stride=1, act=activation)
        self.m = nn.ModuleList(
            [nn.MaxPool2d(kernel_size=ks, stride=1, padding=ks // 2) for ks in kernel_sizes]
        )
        conv2_channels = hidden_channels * (len(kernel_sizes) + 1)
        self.conv2 = BaseConv(conv2_channels, out_channels, 1, stride=1, act=activation)

    def forward(self, x):
        x = self.conv1(x)
        x = torch.cat([x] + [m(x) for m in self.m], dim=1)
        x = self.conv2(x)
        return x



import os
import sys

EXT_NAME_IMG = ['.jpg', '.jpeg', '.webp', '.bmp', '.png']
EXT_VIDEO_IMG = ['mp4', 'mov', 'avi', 'mkv']




def get_path_root():
    '''os.path.dirname(os.path.abspath(__file__))'''
    debug_vars = dict((a, b) for a, b in os.environ.items() if a.find('IPYTHONENABLE') >= 0)
    # 根据不同场景获取根目录
    if len(debug_vars) > 0:
        """当前为debug运行时"""
        path_root = sys.path[2]
    elif getattr(sys, 'frozen', False):
        """当前为exe运行时"""
        path_root = os.getcwd()
    else:
        """正常执行"""
        path_root = sys.path[1]
    path_root = path_root.replace("\\", "/")  # 替换斜杠
    return path_root


def get_img_file(path):
    files_pic = []
    for maindir, subdir, file_name_list in os.walk(path):
        for filename in file_name_list:
            apath = os.path.join(maindir, filename)
            ext = os.path.splitext(apath)[1]
            if ext in EXT_NAME_IMG:
                files_pic.append(apath)
    return files_pic


def fshow_time(fun, arg_list=None, num_time=1):
    '''

    :param fun:
    :param arg_list:  net, input  arg 需在外面打包为list
    :return:
    '''
    import time
    from ftools.GLOBAL_LOG import flog
    assert num_time >= 1
    ret = None
    start = time.time()
    flog.debug('show_time---开始---%s-------' % (fun.__name__))
    for i in range(num_time):
        if arg_list is None:
            ret = fun()
        else:
            ret = fun(*arg_list)
    flog.debug('show_time---完成---%s---%s\n' % (fun.__name__, time.time() - start))
    return ret



import numpy as np
import random
import cv2


def non_max_suppress(predicts_dict, threshold):
    for object_name, bbox in predicts_dict.items():  # 对每一个类别分别进行NMS；一次读取一对键值（即某个类别的所有框）
        bbox_array = np.array(bbox, dtype=np.float)
        # 下面分别获取框的左上角坐标（x1，y1），右下角坐标（x2，y2）及此框的置信度；这里需要注意的是图像左上角可以看做坐标点（0,0），右下角可以看做坐标点（1,1），也就是说从左往右x值增大，从上往下y值增大
        x1 = bbox_array[:, 0]
        y1 = bbox_array[:, 1]
        x2 = bbox_array[:, 2]
        y2 = bbox_array[:, 3]
        scores = bbox_array[:, 4]
        order = scores.argsort()[::-1]  # argsort函数返回的是数组值从小到大的索引值,[::-1]表示取反。即这里返回的是数组值从大到小的索引值
        areas = (x2 - x1 + 1) * (
                y2 - y1 + 1)  # 当前类所有框的面积(python会自动使用广播机制，相当于MATLAB中的.*即两矩阵对应元素相乘)；x1=3,x2=5,习惯上计算x方向长度就是x=3、4、5这三个像素，即5-3+1=3，而不是5-3=2，所以需要加1
        keep = []

        # 按confidence从高到低遍历bbx，移除所有与该矩形框的IoU值大于threshold的矩形框
        while order.size > 0:
            i = order[0]
            keep.append(i)  # 保留当前最大confidence对应的bbx索引
            # 获取所有与当前bbx的交集对应的左上角和右下角坐标，并计算IoU（注意这里是同时计算一个bbx与其他所有bbx的IoU）

            xx1 = np.maximum(x1[i], x1[order[1:]])  # 最大置信度的左上角坐标分别与剩余所有的框的左上角坐标进行比较，分别保存较大值；因此这里的xx1的维数应该是当前类的框的个数减1
            yy1 = np.maximum(y1[i], y1[order[1:]])
            xx2 = np.minimum(x2[i], x2[order[1:]])
            yy2 = np.minimum(y2[i], y2[order[1:]])
            inter = np.maximum(0.0, xx2 - xx1 + 1) * np.maximum(0.0, yy2 - yy1 + 1)
            iou = inter / (areas[i] + areas[order[1:]] - inter)  # 注意这里都是采用广播机制，同时计算了置信度最高的框与其余框的IoU
            inds = np.where(iou <= threshold)[0]  # 保留iou小于等于阙值的框的索引值
            order = order[inds + 1]  # 将order中的第inds+1处的值重新赋值给order；即更新保留下来的索引，加1是因为因为没有计算与自身的IOU，所以索引相差１，需要加上
        bbox = bbox_array[keep]
        predicts_dict[object_name] = bbox.tolist()
    # predicts_dict = predicts_dict
    return predicts_dict


if __name__ == '__main__':
    # 下面在一张全黑图片上测试非极大值抑制的效果
    img = np.zeros((600, 600), np.uint8)
    # predicts_dict = {'black1': [[83, 54, 165, 163, 0.8], [67, 48, 118, 132, 0.5], [91, 38, 192, 171, 0.6]]}
    predicts_dict = {'black1': [[83, 54, 165, 163, 0.8], [67, 48, 118, 132, 0.5], [91, 38, 192, 171, 0.6]],
                     'black2': [[59, 120, 137, 368, 0.12], [54, 154, 148, 382, 0.13]]}
    # 在全黑的图像上画出设定的几个框
    for object_name, bbox in predicts_dict.items():
        for box in bbox:
            x1, y1, x2, y2, score = box[0], box[1], box[2], box[3], box[-1]
            y_text = int(random.uniform(y1,
                                        y2))  # uniform()是不能直接访问的，需要导入 random 模块，然后通过 random 静态对象调用该方法。uniform() 方法将随机生成下一个实数，它在 [x, y) 范围内
            cv2.rectangle(img, (x1, y1), (x2, y2), (255, 255, 255), 2)
            cv2.putText(img, str(score), (x2 - 30, y_text), 2, 1, (255, 255, 0))
        cv2.namedWindow("black1_roi")  # 创建一个显示图像的窗口
        cv2.imshow("black1_roi", img)  # 在窗口中显示图像;注意这里的窗口名字如果不是刚刚创建的窗口的名字则会自动创建一个新的窗口并将图像显示在这个窗口
        cv2.waitKey(0)  # 如果不添这一句，在IDLE中执行窗口直接无响应。在命令行中执行的话，则是一闪而过。
        cv2.destroyAllWindows()  # 最后释放窗口是个好习惯！

    # 在全黑图片上画出经过非极大值抑制后的框
    img_cp = np.zeros((600, 600), np.uint8)
    predicts_dict_nms = non_max_suppress(predicts_dict, 0.1)
    for object_name, bbox in predicts_dict_nms.items():
        for box in bbox:
            x1, y1, x2, y2, score = int(box[0]), int(box[1]), int(box[2]), int(box[3]), box[-1]
            y_text = int(random.uniform(y1,
                                        y2))  # uniform()是不能直接访问的，需要导入 random 模块，然后通过 random 静态对象调用该方法。uniform() 方法将随机生成下一个实数，它在 [x, y) 范围内
            cv2.rectangle(img_cp, (x1, y1), (x2, y2), (255, 255, 255), 2)
            cv2.putText(img_cp, str(score), (x2 - 30, y_text), 2, 1, (255, 255, 0))
        cv2.namedWindow("black1_nms")  # 创建一个显示图像的窗口
        cv2.imshow("black1_nms", img_cp)  # 在窗口中显示图像;注意这里的窗口名字如果不是刚刚创建的窗口的名字则会自动创建一个新的窗口并将图像显示在这个窗口
        cv2.waitKey(0)  # 如果不添这一句，在IDLE中执行窗口直接无响应。在命令行中执行的话，则是一闪而过。
        cv2.destroyAllWindows()  # 最后释放窗口是个好习惯！



import sys

import torch

from ftools.GLOBAL_LOG import flog
from ftools.ai.fits.fweight import save_weight


def init_weight4nanodet(model, device, path_host):
    '''
    421个
    'model.head.distribution_project.project'
    'model.fpn.lateral_convs.0.conv.weight'
    'model.head.gfl_reg.2.bias'
    'model.head.cls_convs.2.1.pointwise.weight'
    '''
    file_weight = path_host + '/AI/weights/nanodet/nanodet_m.ckpt'
    # file_weight = 'M:/AI/weights/feadre/nanodet_50_57.765.pth'
    checkpoint = torch.load(file_weight, map_location=device)
    pretrained_dict_y = checkpoint['state_dict']
    s_backbone = 'net.backbone.'  # lrelu
    # s_backbone = 'net.backbone.model_hook.'  # relu
    s_fpn = ['net.fpn.prj_5.', 'net.fpn.prj_4.', 'net.fpn.prj_3.', ]
    # s_fpn = ['net.fpn.lateral_convs.2.conv.', 'net.fpn.lateral_convs.1.conv.', 'net.fpn.lateral_convs.0.conv.', ]
    s_gfl_head_convs_front = 'net.gfl_head.convs_front.'
    s_gfl_head_cls_reg = 'net.gfl_head.cls_reg.'
    dd = {}
    for k, v in pretrained_dict_y.items():
        split_key = k.split(".")
        if 'backbone' in split_key:
            det = '.'.join(split_key[2:])
            dd[s_backbone + det] = v
        elif 'fpn' in split_key:
            det = '.'.join(split_key[-1:])
            if '0' in split_key:
                dd[s_fpn[2] + det] = v
            elif '1' in split_key:
                dd[s_fpn[1] + det] = v
            elif '2' in split_key:
                dd[s_fpn[0] + det] = v
        elif 'head' in split_key:
            if 'distribution_project' in split_key:
                # 这个是 tensor([0., 1., 2., 3., 4., 5., 6., 7.]) 没用
                pass
            elif 'cls_convs' in split_key:
                det = '.'.join(split_key[3:])
                dd[s_gfl_head_convs_front + det] = v
            elif 'gfl_cls' in split_key:  # 迁移学习改这里
                det = '.'.join(split_key[-2:])
                dd[s_gfl_head_cls_reg + det] = v
            elif 'gfl_reg' in split_key:
                pass

    keys_missing, keys_unexpected = model.load_state_dict(dd, strict=False)
    if len(keys_missing) > 0 or len(keys_unexpected):
        flog.error('missing_keys %s', keys_missing)  # 这个是 model 的属性
        flog.error('unexpected_keys %s', keys_unexpected)  # 这个是 pth 的属性
    flog.debug('权重加载完成 %s', file_weight)

    # path_root_weight = r'M:\AI\weights\feadre\zz'
    # save_weight(path_root_weight, model=model, name='nanodet_coco_p34r29')
    sys.exit(-1)


def init_weight4yolox(model, device, path_host):
    '''

    '''
    file_weight = path_host + '/AI/weights/yolox/yolox_nano.pth.tar'
    checkpoint = torch.load(file_weight, map_location=device)
    start_epoch = 1
    if 'start_epoch' in checkpoint:
        start_epoch = checkpoint['start_epoch'] + 1
    pretrained_dict_y = checkpoint['model']

    s_backbone = 'net.'
    s_head = s_backbone
    dd = {}
    for k, v in pretrained_dict_y.items():
        split_key = k.split(".")
        if 'backbone' in split_key:
            det = '.'.join(split_key[1:])
            dd[s_backbone + det] = v
        elif 'head' in split_key:
            dd[s_head + k] = v
        else:
            dd[k] = v

    keys_missing, keys_unexpected = model.load_state_dict(dd, strict=False)
    if len(keys_missing) > 0 or len(keys_unexpected):
        flog.error('missing_keys %s', keys_missing)  # 这个是 model 的属性
        print()
        flog.error('unexpected_keys %s', keys_unexpected)  # 这个是 pth 的属性
    flog.debug('权重加载完成 %s', file_weight)

    path_root_weight = r'M:\AI\weights\feadre\zz'
    save_weight(path_root_weight, epoch=start_epoch,
                model=model, name='yolox_coco_pxxrxx_y')
    sys.exit(-1)



import numpy as np


def fcre_kps_min_bbox(kps, scale=1):
    '''
    创建 关键点最小外拉框
    :param kps: (ngt,num_keypoints*3)
    :param scale:
    :return:
    '''
    _kps, mask = fsplit_kps(kps)
    xys = _kps[:, :, :2]
    bbox = []
    for i, xy in enumerate(xys):
        lt = np.min(xy[mask[i]], axis=0)
        rb = np.max(xy[mask[i]], axis=0)
        bbox.append([*lt, *rb])
    bbox = np.stack(bbox, 0)
    return bbox * scale


def fverify_kps(kps, ltrb):
    '''
    校验有效性 出界的不要
    :param kps: (ngt,num_keypoints*3)
    :param ltrb: [4]
    :return:
    '''
    assert len(ltrb) == 4, 'len(ltrb)= %s' % len(ltrb)
    ngt, dim_kps = kps.shape
    # (ngt,dim_kps) ->  (ngt,num_keypoints,3)
    _kps = kps.reshape(ngt, -1, 3)
    # (ngt,num_keypoints) 超出的记录下来
    mask_x = np.logical_or(_kps[:, :, 0] < ltrb[0], _kps[:, :, 0] > ltrb[2])
    mask_y = np.logical_or(_kps[:, :, 1] < ltrb[1], _kps[:, :, 1] > ltrb[3])
    mask = np.logical_or(mask_x, mask_y)
    if mask.any():
        _kps[mask, 2] = 0
        # flog.debug('mask.sum()= %s', mask.sum())
    return _kps.reshape(ngt, -1)


def fsplit_kps(kps):
    ''' 返回 reshape后的kps 和 2二维 mask '''
    ngt, dim_kps = kps.shape
    # (ngt,dim_kps) ->  (ngt,num_keypoints,3)
    _kps = kps.reshape(ngt, -1, 3)
    # xys = _kps[:, :, :2]
    # (ngt,num_keypoints)
    mask = _kps[:, :, 2].astype(np.bool)
    return _kps, mask



import math
import os

import torch
import cv2
import numpy as np
from numpy import random

from ftools.GLOBAL_LOG import flog
from ftools.ai.object_detection.boxes.f_boxes import ltwh2ltrb, fverify_bbox
from ftools.ai.datas.dataset_coco import CustomCocoDataset
from ftools.ai.object_detection.boxes.f_kps import fcre_kps_min_bbox, fverify_kps, fsplit_kps
from ftools.ai.picture.f_size_handler import resize_np_keep
from ftools.ai.picture.f_show import f_show_od_np4plt_v3, f_show_kp_np4cv, f_show_od_np4cv
from ftools.ai.datas.z_dataloader import KEYPOINTS_Mirror


def ftransform_nanodet_train(size_wh_input):
    return FCompose([
        Uint2Float32_np(),
        FResizeKeep(size_wh_input),
        RandomBrightness(50),
        RandomContrast(lower=0.8, upper=1.2),
        ConvertColor(current='BGR', transform='HSV'),  # bgr -> hsv
        RandomSaturation(lower=0.8, upper=1.2),  # 随机色彩'
        ConvertColor(current='HSV', transform='BGR'),  # hsv -> bgr
        FRandomMirror(),
        Normalize_v2(),
        to_tensor_target2one(is_box_oned=False),
    ])


def ftransform_nanodet_test(size_wh_input):
    return FCompose([
        FResizeKeep(size_wh_input),
        Normalize_v2(),
        to_tensor_target2one(is_box_oned=False),
    ])


# -----------------------------------------------------------
def ftransform_yolox_train4mosaic(size_wh_input):
    return FCompose([
        Uint2Float32_np(),
        RandomContrast_v2(),
        ConvertColor(current='BGR', transform='HSV'),
        FRandomHSV(),
        RandomSaturation(),
        ConvertColor(current='HSV', transform='BGR'),
        FRandomMirror(),
        Normalize_v2(),
        ConvertColor(current='BGR', transform='RGB'),
        to_tensor_target2one(is_box_oned=False),
    ])


def ftransform_yolox_train(size_wh_input):
    return FCompose([
        Uint2Float32_np(),
        RandomContrast_v2(),
        ConvertColor(current='BGR', transform='HSV'),
        FRandomHSV(),
        RandomSaturation(),
        ConvertColor(current='HSV', transform='BGR'),
        FRandomMirror(),
        FResizeKeep(size_wh_input, fill_color=(128, 128, 128)),
        Normalize_v2(),
        ConvertColor(current='BGR', transform='RGB'),
        to_tensor_target2one(is_box_oned=False),
    ])


def ftransform_yolox_test(size_wh_input):
    return FCompose([
        FResizeKeep(size_wh_input, fill_color=(128, 128, 128)),
        Normalize_v2(),
        ConvertColor(current='BGR', transform='RGB'),
        to_tensor_target2one(is_box_oned=False),
    ])


# -----------------------------------------------------------
def ftransform_ssd_train(size_wh_input, is_multi_scale=None, multi_scale_val=None):
    return FCompose([
        Uint2Float32_np(),  # image int8转换成float [0,256)
        # --------------- 图形整体 ------------
        PhotometricDistort(),
        # --------------- 变换 ------------
        FExpand(),  # 放大缩小图片 只会缩小
        FRandomSampleCrop(),  # 随机剪切定位 keypoints 只会放大
        FRandomMirror(),  # 水平镜像
        FResize(size_wh_input, is_multi_scale, multi_scale_val),

        # --------------- 后处理 ------------
        Normalize(),
        ConvertColor(current='BGR', transform='RGB'),
        to_tensor_target2one(is_box_oned=False),
    ])


def ftransform_cpm_train(size_wh_input, is_multi_scale=None, multi_scale_val=None):
    return FCompose([
        FRandomRotate(40),
        FRandomSampleCrop(),
        FRandomMirror(KEYPOINTS_Mirror['lsp']),
        FResize(size_wh_input, is_multi_scale, multi_scale_val),
        Normalize(),
        ConvertColor(current='BGR', transform='RGB'),
        to_tensor_target2one(is_box_oned=False),
    ])


def ftransform_ssd_test(size_wh_input, is_multi_scale=None, multi_scale_val=None):
    return FCompose([
        FResize(size_wh_input, is_multi_scale, multi_scale_val),
        Normalize(),
        ConvertColor(current='BGR', transform='RGB'),
        to_tensor_target2one(is_box_oned=False),
    ])


''' ------------------ 以上是套装 ----------------------  '''


def _copy_box(boxes):
    if isinstance(boxes, np.ndarray):
        boxes_ = boxes.copy()
    elif isinstance(boxes, torch.Tensor):
        boxes_ = boxes.clone()
    else:
        raise Exception('类型错误', type(boxes))

    return boxes_


def intersect(box_a, box_b):
    max_xy = np.minimum(box_a[:, 2:], box_b[2:])
    min_xy = np.maximum(box_a[:, :2], box_b[:2])
    inter = np.clip((max_xy - min_xy), a_min=0, a_max=np.inf)
    return inter[:, 0] * inter[:, 1]


def jaccard_numpy(box_a, box_b):
    """Compute the jaccard overlap of two sets of boxes.  The jaccard overlap
    is simply the intersection over union of two boxes.
    E.g.:
        A ∩ B / A ∪ B = A ∩ B / (area(A) + area(B) - A ∩ B)
    Args:
        box_a: Multiple bounding boxes, Shape: [num_boxes,4]
        box_b: Single bounding box, Shape: [4]
    Return:
        jaccard overlap: Shape: [box_a.shape[0], box_a.shape[1]]
    """
    inter = intersect(box_a, box_b)
    area_a = ((box_a[:, 2] - box_a[:, 0]) *
              (box_a[:, 3] - box_a[:, 1]))  # [A,B]
    area_b = ((box_b[2] - box_b[0]) *
              (box_b[3] - box_b[1]))  # [A,B]
    union = area_a + area_b - inter
    return inter / union  # [A,B]


class FCompose:

    def __init__(self, transforms):
        super(FCompose, self).__init__()
        self.transforms = transforms

    def __call__(self, img, target={}):
        # f_plt_show_cv(image,boxes)
        for t in self.transforms:
            img, target = t(img, target)
            if target is not None:
                if 'boxes' in target and 'labels' in target:
                    if len(target['boxes']) != len(target['labels']):
                        flog.warning('!!! 数据有问题 Compose  %s %s %s ', len(target['boxes']), len(target['labels']), t)
        return img, target


class Uint2Float32_np(object):
    ''' 在处理开始时使用 '''

    def __call__(self, image, target):
        ''' cv打开的np 默认是uint8 '''
        return image.astype(np.float32), target


class ts2img_np_bgr(object):
    def __call__(self, tensor, target):
        return tensor.cpu().numpy().astype(np.float32).transpose((1, 2, 0)), target


# --------------- 图形整体 ------------

class FRandomNoise:
    '''
    随机噪声 噪点
    :param image:
    :param threshold:
    :return:
    '''

    def __init__(self, threshold=32) -> None:
        super().__init__()
        self.threshold = threshold

    def __call__(self, img_np, target):
        if random.randint(2):
            noise = np.random.uniform(low=-1, high=1, size=img_np.shape)
            img_np = img_np + noise * self.threshold
            img_np = np.clip(img_np, 0, 255)

        return img_np, target


class Ffilter_gaussian:
    '''
    高斯滤波器是根据高斯函数的形状来选择权值的线性平滑滤波器，滤波器符合二维高斯分布
    模糊
    :param image:
    :param threshold:
    :return:
    '''

    def __init__(self, k_size=5) -> None:
        super().__init__()
        self.k_size = k_size

    def __call__(self, img_np, target):
        if random.randint(2):
            img_np = cv2.GaussianBlur(img_np, (self.k_size, self.k_size), 3)
        return img_np, target


class Fgamma_transform:
    '''
    Gamma变换就是用来图像增强，通过非线性变换提升了暗部细节
    :param image:
    :param threshold:
    :return:
    '''

    def __init__(self, gamma=1.6) -> None:
        super().__init__()
        self.gamma = gamma

    def __call__(self, img_np, target):
        if random.randint(2):
            max_value = np.max(img_np)
            min_value = np.min(img_np)
            value_l = max_value - min_value
            img_np = (img_np - min_value) / value_l
            img_np = np.power(img_np, self.gamma)
            img_np = img_np * 255
        return img_np, target


class RandomSaturation(object):
    '''随机色彩 需要HSV'''

    def __init__(self, lower=0.5, upper=1.5):
        self.lower = lower
        self.upper = upper
        assert self.upper >= self.lower, "contrast upper must be >= lower."
        assert self.lower >= 0, "contrast lower must be non-negative."

    def __call__(self, image, target):
        if random.randint(2):
            image[:, :, 1] *= random.uniform(self.lower, self.upper)
        return image, target


class FRandomHSV:
    '''随机色彩 需要HSV'''

    def __init__(self, lower=0.5, upper=1.5):
        self.lower = lower
        self.upper = upper

    def __call__(self, image, target):
        if random.randint(2):
            tmp = image[:, :, 0].astype(int) + random.randint(-18, 18)
            tmp %= 180
            image[:, :, 0] = tmp

        return image, target


class RandomHue(object):
    def __init__(self, delta=18.0):
        assert delta >= 0.0 and delta <= 360.0
        self.delta = delta

    def __call__(self, image, target):
        if random.randint(2):
            image[:, :, 0] += random.uniform(-self.delta, self.delta)
            image[:, :, 0][image[:, :, 0] > 360.0] -= 360.0
            image[:, :, 0][image[:, :, 0] < 0.0] += 360.0
        return image, target


class RandomContrast(object):
    '''随机透明度'''

    def __init__(self, lower=0.5, upper=1.5):
        self.lower = lower
        self.upper = upper
        assert self.upper >= self.lower, "contrast upper must be >= lower."
        assert self.lower >= 0, "contrast lower must be non-negative."

    # expects float image
    def __call__(self, image, target):
        if random.randint(2):  # 50%
            alpha = random.uniform(self.lower, self.upper)
            image *= alpha
            np.clip(image, a_min=0, a_max=255, out=image)
        # f_plt_show_cv(image)
        return image, target


class RandomContrast_v2():
    '''随机透明度'''

    def __init__(self):
        pass

    # expects float image
    def __call__(self, image, target):
        if random.randint(2):
            self._convert(image, beta=random.uniform(-32, 32))
        if random.randint(2):
            self._convert(image, beta=random.uniform(0.5, 1.5))
        # f_plt_show_cv(image)
        return image, target

    def _convert(self, image, alpha=1, beta=0):
        image = image.astype(float) * alpha + beta
        np.clip(image, a_min=0, a_max=255, out=image)
        return image


class RandomGray:
    '''随机灰度'''

    def __call__(self, image, target):
        if random.random() > 0.9:
            _img = np.zeros(image.shape, dtype=np.uint8)
            gray = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
            _img[:, :, 0] = gray
            _img[:, :, 1] = gray
            _img[:, :, 2] = gray
            image = _img

        return image, target


class RandomBrightness(object):
    def __init__(self, delta=32):
        assert delta >= 0.0
        assert delta <= 255.0
        self.delta = delta

    def __call__(self, image, target):
        '''随机亮度增强'''
        if random.randint(2):  # 50%
            delta = random.uniform(-self.delta, self.delta)
            image += delta
        # f_plt_show_cv(image)
        return image, target


# --------------- 变换 ------------
class FRandomRotate:
    '''
    有框不能用旋转
    '''

    def __init__(self, max_degree=20):
        self.max_degree = max_degree

    def __call__(self, img_np, target):
        degree = random.uniform(-self.max_degree, self.max_degree)
        img_h, img_w, c = img_np.shape
        cx, cy = img_w / 2.0, img_h / 2.0

        ''' 图片处理 '''
        # mat rotate 1 center 2 angle 3 缩放系数 (2, 3)
        rotate_matrix = cv2.getRotationMatrix2D((cx, cy), degree, 1.0)
        img_np = cv2.warpAffine(img_np, rotate_matrix, (img_w, img_h),
                                borderValue=(128, 128, 128))
        # debug
        # f_show_od_np4cv(img_np)

        if 'boxes' in target:
            boxes_new = []
            boxes_ltrb = target['boxes']
            for bbox in boxes_ltrb:
                a = rotate_matrix[:, :2]  # a.shape (2,2)
                b = rotate_matrix[:, 2:]  # b.shape(2,1)
                b = np.reshape(b, newshape=(1, 2))
                a = np.transpose(a)

                [left, up, right, down] = bbox
                corner_point = np.array([[left, up], [right, up], [left, down], [right, down]])
                corner_point = np.dot(corner_point, a) + b
                min_left = np.min(corner_point[:, 0])
                min_up = np.min(corner_point[:, 1])
                max_right = np.max(corner_point[:, 0])
                max_down = np.max(corner_point[:, 1])
                boxes_new.append(np.array([min_left, min_up, max_right, max_down]))
            target['boxes'] = np.stack(boxes_new, 0)
            np.clip(target['boxes'][:, ::2], a_min=0, a_max=img_w, out=target['boxes'][:, ::2])
            np.clip(target['boxes'][:, 1::2], a_min=0, a_max=img_h, out=target['boxes'][:, 1::2])

            mask_pos = fverify_bbox(target['boxes'])
            target['boxes'] = target['boxes'][mask_pos]
            if 'labels' in target:
                target['labels'] = target['labels'][mask_pos]

        if 'kps' in target:
            kps = target['kps'].reshape(target['kps'].shape[0], -1, 3)
            for i, kp in enumerate(kps):
                for j, xyt in enumerate(kp):
                    if xyt[2] == 0:
                        continue
                    p = np.array([xyt[0], xyt[1], 1])
                    p = rotate_matrix.dot(p)

                    # 越界判断
                    if p[0] > img_w or p[0] < 0 or p[1] > img_h or p[1] < 0:
                        xyt[2] = 0

                    xyt[0] = p[0]
                    xyt[1] = p[1]
                    kps[i, j] = xyt

            # 共享内存不用赋值
            # target['kps'] = kps.reshape(target['kps'].shape[0], -1)
            # debug
            # f_show_kp_np4cv(img_np, kps=target['kps'],
            #                 kps_seq=kps_seq)
        return img_np, target


class FExpand:
    '''随机 缩小放在图片中某处  其它部份为黑色'''

    def __init__(self, mean=(128, 128, 128), num_keypoints=None):
        ''' 均值用于扩展边界 '''
        self.mean = mean  # 这个用于背景
        self.num_keypoints = num_keypoints

    def __call__(self, image, target):
        if random.randint(2):
            return image, target

        height, width, depth = image.shape
        ratio = random.uniform(1, 3)  # 缩小 1~4的一个比例  这个必须大于1 只能缩小
        left = random.uniform(0, width * ratio - width)
        top = random.uniform(0, height * ratio - height)

        expand_image = np.zeros(
            (int(height * ratio), int(width * ratio), depth),
            dtype=image.dtype)
        expand_image[:, :, :] = self.mean
        expand_image[int(top):int(top + height), int(left):int(left + width)] = image
        image = expand_image

        if target is None:
            return image, target

        # 这个只缩小不存在越界问题
        if 'boxes' in target:
            # 原尺寸不要了
            boxes = target['boxes']
            if isinstance(boxes, np.ndarray):
                boxes[:, :2] += (int(left), int(top))  # lt
                boxes[:, 2:] += (int(left), int(top))  # rb
            elif isinstance(boxes, torch.Tensor):
                boxes[:, :2] += torch.tensor((int(left), int(top)))
                boxes[:, 2:] += torch.tensor((int(left), int(top)))
            else:
                raise Exception('target[boxes]类型错误', type(boxes))

        if 'kps' in target:
            _kps, mask = fsplit_kps(target['kps'])
            _kps[..., :2] += (int(left), int(top))

        # f_plt_show_cv(image,torch.tensor(boxes))
        return image, target


class FRandomSampleCrop:
    """
    随机Crop
    """

    def __init__(self, num_keypoints=None):
        # 随机源
        self.sample_options = (
            # using entire original input image
            None,
            # sample a patch s.t. MIN jaccard w/ obj in .1,.3,.4,.7,.9
            (0.1, None),
            (0.3, None),
            (0.7, None),
            (0.9, None),
            # randomly sample a patch
            (None, None),
        )
        self.num_keypoints = num_keypoints

    def __call__(self, image, target):
        height, width, _ = image.shape
        # bbox = fcre_kps_min_bbox(target['kps'])
        # f_show_kp_np4cv(image,
        #                 kps=target['kps'],
        #                 kps_seq=target['kps_seq'],
        #                 g_ltrb=bbox,
        #                 )
        while True:
            # randomly choose a mode
            mode = random.choice(self.sample_options)
            if mode is None:
                return image, target

            min_iou, max_iou = mode
            if min_iou is None:
                min_iou = float('-inf')
            if max_iou is None:
                max_iou = float('inf')

            # max trails (50) 找一个 iou在范围内 且 保留BOX中心的剪切
            for _ in range(50):

                w = random.uniform(0.3 * width, width)
                h = random.uniform(0.3 * height, height)

                # aspect ratio constraint b/t .5 & 2
                if h / w < 0.5 or h / w > 2:
                    continue

                left = random.uniform(width - w)
                top = random.uniform(height - h)

                # convert to integer rect x1,y1,x2,y2
                rect = np.array([int(left), int(top), int(left + w), int(top + h)])

                # calculate IoU (jaccard overlap) b/t the cropped and gt boxes
                if 'boxes' in target:
                    bbox = target['boxes']
                    overlap = jaccard_numpy(bbox, rect)
                elif 'kps' in target:
                    # 外接框
                    bbox = fcre_kps_min_bbox(target['kps'])
                    overlap = jaccard_numpy(bbox, rect)
                else:
                    raise Exception('boxes 和 kps 不存在,无法使用 FRandomSampleCrop')

                # is min and max overlap constraint satisfied? if not try again
                if len(overlap) == 0 or (overlap.min() < min_iou and max_iou < overlap.max()):
                    continue

                # 只保留中心点在剪切框内的 cut the crop from the image
                img_new = image[rect[1]:rect[3], rect[0]:rect[2], :]

                # keep overlap with gt box IF center in sampled patch
                centers = (bbox[:, :2] + bbox[:, 2:]) / 2.0
                # --- 是否切到box 需修正
                # mask in all gt boxes that above and to the left of centers
                m1 = (rect[0] < centers[:, 0]) * (rect[1] < centers[:, 1])
                # mask in all gt boxes that under and to the right of centers
                m2 = (rect[2] > centers[:, 0]) * (rect[3] > centers[:, 1])
                # mask in that both m1 and m2 are true
                mask = m1 * m2

                #  have any valid boxes? try again if not
                if not mask.any():  # 一个都没有
                    continue

                # 关键点需要直接修复
                if 'kps' in target:
                    target['kps'] = fverify_kps(target['kps'], rect)
                    # 共享内存
                    _kps, _mask = fsplit_kps(target['kps'])
                    _kps[:, :, :2] = _kps[:, :, :2] - rect[:2][None, None, :]
                    # raise Exception('暂不支持')

                # take only matching gt labels
                if 'boxes' in target:
                    # 取出要修复的
                    current_boxes = bbox[mask, :]
                    # should we use the box left and top corner or the crop's
                    current_boxes[:, :2] = np.maximum(current_boxes[:, :2], rect[:2])
                    # adjust to crop (by substracting crop's left,top)
                    current_boxes[:, :2] -= rect[:2]

                    current_boxes[:, 2:] = np.minimum(current_boxes[:, 2:], rect[2:])
                    # adjust to crop (by substracting crop's left,top)
                    current_boxes[:, 2:] -= rect[:2]

                    target['boxes'] = current_boxes

                    # 校验有效性
                    mask_pos = fverify_bbox(target['boxes'])
                    target['boxes'] = target['boxes'][mask_pos]
                    if 'labels' in target:
                        target['labels'] = target['labels'][mask][mask_pos]

                    # debug
                    # f_show_od_np4cv(img_new, current_boxes)

                # debug
                # f_show_kp_np4cv(img_new,
                #                 kps=target['kps'],
                #                 kps_seq=None,
                #                 g_ltrb=target['boxes'],
                #                 is_color_same=True,
                #                 )
                return img_new, target


class FRandomMirror:
    '''随机水平镜像'''

    def __init__(self, rules_flip=None) -> None:
        super().__init__()
        self.rules_flip = rules_flip

    def __call__(self, image, target):
        _, width, _ = image.shape
        if random.randint(2):
            # flog.debug('FRandomMirror %s', )
            image = image[:, ::-1]
            # boxes = boxes.copy()

            if target is None:
                return image, target

            if 'boxes' in target:
                target['boxes'][:, [0, 2]] = width - target['boxes'][:, [2, 0]]

            if 'kps' in target:
                ''' 这个需要订制一个翻转规则  '''
                assert self.rules_flip is not None, 'RandomMirror keypoints 不支持 需传入 rules_flip '
                # 批量修改x值 y不变
                target['kps'][:, ::3] = width - target['kps'][:, ::3]

                _kps, mask = fsplit_kps(target['kps'])

                for xyt in _kps:
                    # 改变 kk可以改变 target['kps'] 共享内存
                    for s, d in self.rules_flip:
                        _t = xyt[s].copy()
                        xyt[s] = xyt[d]
                        xyt[d] = _t
                    pass
                # 共享内存直接调整 无需这个
                # target['kps'] = kks.reshape(n, -1)

            # target['kps'].view(-1,2).flip(0)

        return image, target


class FResizeKeep:
    def __init__(self, size_wh_new, fill_color=(128, 128, 128)):
        '''

        '''
        self.size_wh_new = size_wh_new
        self.fill_color = fill_color

    def __call__(self, img_np, target):
        img_np, ratio, hw, (left, top, right, bottom) = resize_np_keep(
            img_np,
            self.size_wh_new,
            mode='center',
            # mode='lt',
            fill_color=self.fill_color,
        )

        # debug
        # import matplotlib.pyplot as plt
        # plt.imshow(img_np)
        # plt.show()

        if target is not None:
            if 'boxes' in target:
                target['boxes'][:, [0, 2]] = target['boxes'][:, [0, 2]] * ratio
                target['boxes'][:, [1, 3]] = target['boxes'][:, [1, 3]] * ratio
                # target['boxes'] += np.array((left, top, right, bottom), dtype=np.float32)
                target['boxes'] += np.array((left, top, left, top), dtype=np.float32)
            if 'kps' in target:
                # 保持第三维不变
                num_keypoints = target['kps'].shape[1] // 3
                target['kps'] = target['kps'] \
                                * np.tile(np.array([ratio] * 2 + [1])[None, :], (1, num_keypoints))
                target['kps'] += np.tile(np.array([left, top, 0])[None, :], (1, num_keypoints))
                # debug
                # f_show_kp_np4cv(img_np, kps=target['kps'],
                #                 kps_seq=target['kps_seq'])

            # 修正偏移
            target['off_ltrb'] = (left, top, right, bottom)

        return img_np, target


class FResize:
    def __init__(self, size_wh=None, is_multi_scale=False, multi_scale_val=(800, 1333)):
        '''
        直接拉伸和 多尺度
        '''
        self.size_wh = size_wh
        self.is_multi_scale = is_multi_scale
        self.multi_scale_val = multi_scale_val

    def __call__(self, img_np, target):
        if self.is_multi_scale:
            min_side, max_side = self.multi_scale_val
            h, w, _ = img_np.shape
            smallest_side = min(w, h)
            largest_side = max(w, h)
            scale = min_side / smallest_side
            if largest_side * scale > max_side:
                scale = max_side / largest_side
            nw, nh = int(scale * w), int(scale * h)
            image_resized = cv2.resize(img_np, (nw, nh))

            if target is not None:
                target['toone'] = image_resized.shape[:2][::-1]
            pad_w = 32 - nw % 32
            pad_h = 32 - nh % 32
            # 右下加上pad 不影响targets
            image_res = np.zeros(shape=[nh + pad_h, nw + pad_w, 3], dtype=np.uint8)
            image_res[:nh, :nw, :] = image_resized

            scale_w = scale
            scale_h = scale
        else:
            assert self.size_wh is not None
            size_wh_z = self.size_wh
            w_ratio, h_ratio = np.array(img_np.shape[:2][::-1]) / np.array(size_wh_z)

            image_res = cv2.resize(img_np, (size_wh_z[0], size_wh_z[1]))

            scale_w = 1 / w_ratio
            scale_h = 1 / h_ratio

        if target is not None:
            if 'boxes' in target:
                target['boxes'][:, [0, 2]] = target['boxes'][:, [0, 2]] * scale_w
                target['boxes'][:, [1, 3]] = target['boxes'][:, [1, 3]] * scale_h
            if 'kps' in target:
                # 保持第三维不变
                num_keypoints = target['kps'].shape[1] // 3
                target['kps'] = target['kps'] \
                                * np.tile(np.array([scale_w, scale_h, 1])[None, :], (1, num_keypoints))
                # # debug
                # f_show_kp_np4cv(image_res, kps=target['kps'],
                #                 kps_seq=kps_seq)

        return image_res, target


# --------------- 后处理 ------------
class Normalize:
    # def __init__(self, mean=(0.406, 0.456, 0.485), std=(0.225, 0.224, 0.229)):
    def __init__(self, mean=(128.0, 128.0, 128.0), std=(256.0, 256.0, 256.0)):
        self.mean = np.array(mean, dtype=np.float32)
        self.std = np.array(std, dtype=np.float32)

    def __call__(self, image, target):
        ''' -0.5 ~ 0.5 '''
        # bgr
        # f_show_pic_np4plt(image)

        # (928, 1216, 3)
        image = image.astype(np.float32).clip(0.0, 255.)
        # image /= 255.
        image -= self.mean.reshape(1, 1, -1)
        image /= self.std

        if image.max() > 0.5 or image.min() < -0.5:
            raise Exception('图片有问题 mean = %s ,std = %s, max=%s,min=%s'
                            % (str(self.mean), str(self.std), image.max(), image.min()))

        # f_show_pic_np4plt(image)
        # image = f_recover_normalization4np(image)
        # f_show_pic_np4plt(image)

        return image, target


class Normalize_v2:
    def __init__(self, mean_bgr=(103.53, 116.28, 123.675), std_bgr=(57.375, 57.12, 58.395), is_tooned=False):
        '''

        mean_bgr = (103.53, 116.28, 123.675)
        mean_bgr_one = (103.53, 116.28, 123.675)
        std_bgr = (0.406, 0.456, 0.485)
        std_bgr_one = (0.225, 0.224, 0.229)

        :param mean_bgr:
        :param std_bgr:
        :param is_tooned:
        '''
        self.mean = np.array(mean_bgr, dtype=np.float32)
        self.std = np.array(std_bgr, dtype=np.float32)
        self.is_tooned = is_tooned

    def __call__(self, image, target):
        # f_show_od_np4plt_v3(image, target['boxes'],
        #                     is_recover_size=False
        #                     )

        img = image.astype(np.float32) / 255
        if self.is_tooned:
            mean = np.array(self.mean, dtype=np.float32).reshape(1, 1, 3)
            std = np.array(self.std, dtype=np.float32).reshape(1, 1, 3)
        else:
            # 维度匹配
            mean = np.array(self.mean, dtype=np.float32).reshape(1, 1, 3) / 255
            std = np.array(self.std, dtype=np.float32).reshape(1, 1, 3) / 255
        img = (img - mean) / std

        return img, target


class ConvertColor:
    '''bgr -> hsv'''

    def __init__(self, current='BGR', transform='HSV'):
        self.transform = transform
        self.current = current

    def __call__(self, image, target):
        # plt.imshow(image)
        # plt.show()
        if self.current == 'BGR' and self.transform == 'HSV':
            image = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        elif self.current == 'HSV' and self.transform == 'BGR':
            image = cv2.cvtColor(image, cv2.COLOR_HSV2BGR)
        elif self.current == 'BGR' and self.transform == 'RGB':
            # image = image[:, :, ::-1]
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        else:
            raise NotImplementedError
        return image, target


class to_tensor_target2one:
    def __init__(self, is_box_oned=False) -> None:
        super().__init__()
        self.is_box_oned = is_box_oned

    def __call__(self, cvimage, target):
        if target is not None and self.is_box_oned:
            # np整体复制 wh
            whwh = np.tile(cvimage.shape[:2][::-1], 2)
            if 'boxes' in target:
                target['boxes'][:, :] = target['boxes'][:, :] / whwh

            if 'kps' in target:
                target['kps'][:, ::3] /= whwh[0]
                target['kps'][:, 1::3] /= whwh[1]

        # (h,w,c -> c,h,w) = bgr
        return torch.from_numpy(cvimage.astype(np.float32)).permute(2, 0, 1), target


# ------------------------------ 无用 --------------------------------
class Target2one:
    '''原图转归一化尺寸  ToAbsoluteCoords 相反'''

    def __call__(self, image, target):
        if target is None:
            return image, target
        whwh = np.tile(image.shape[:2][::-1], 2)
        if 'boxes' in target:
            target['boxes'][:, :] = target['boxes'][:, :] / whwh

        if 'kps' in target:
            target['kps'][:, ::3] /= whwh[0]
            target['kps'][:, 1::3] /= whwh[1]
        return image, target


class ToAbsoluteCoords:
    ''' boxes 恢复原图尺寸  归一化尺寸转原图  Target2one 相反'''

    def __call__(self, image, target):
        '''
        归一化 -> 绝对坐标
        :param image:
        :param boxes:
        :param labels:
        :return:
        '''
        if target is None:
            return image, target
        whwh = np.tile(image.shape[:2][::-1], 2)
        if 'boxes' in target:
            target['boxes'][:, :] = target['boxes'][:, :] * whwh

        if 'kps' in target:
            target['kps'][:, ::3] *= whwh[0]
            target['kps'][:, 1::3] *= whwh[1]
        return image, target


class PhotometricDistort(object):
    '''图片增强'''

    def __init__(self):
        self.pd = [
            RandomContrast(),  # 随机透明度
            ConvertColor(transform='HSV'),  # bgr -> hsv
            RandomSaturation(),  # 随机色彩'
            RandomHue(),
            ConvertColor(current='HSV', transform='BGR'),  # hsv -> bgr
            RandomContrast()  # 随机透明度
        ]
        self.rand_brightness = RandomBrightness()  # 随机亮度增强
        # self.rand_light_noise = RandomLightingNoise()  # 颜色杂音

    def __call__(self, image, target):
        im = image.copy()
        im, target = self.rand_brightness(im, target)
        if random.randint(2):  # 先转换还是后转换
            distort = FCompose(self.pd[:-1])
        else:
            distort = FCompose(self.pd[1:])
        im, target = distort(im, target)
        return im, target
    # return self.rand_light_noise(im, boxes, labels)


def f_recover_normalization4ts(img_ts, mean_bgr=(128.0, 128.0, 128.0), std_bgr=(256.0, 256.0, 256.0)):
    '''

    :param img_ts: c,h,w
    :return:
    '''
    device = img_ts.DEVICE
    # torch.Size([3, 928, 1536]) -> torch.Size([928, 1536, 3])
    # img_ts_show = img_ts.permute(1, 2, 0) # rgb -> bgr
    mean_bgr = torch.tensor(mean_bgr, device=device)[:, None, None]  # 3,1,1
    std_bgr = torch.tensor(std_bgr, device=device).unsqueeze(-1).unsqueeze(-1)
    img_ts = img_ts * std_bgr + mean_bgr
    # img_ts_show = img_ts_show.permute(2, 0, 1)# bgr -> rgb
    return img_ts


def f_recover_normalization4ts_v2(img_ts, mean_bgr=(103.53, 116.28, 123.675), std_bgr=(57.375, 57.12, 58.395)):
    '''

    :param img_ts: c,h,w
    :return:
    '''
    device = img_ts.DEVICE
    mean_bgr = (torch.tensor(mean_bgr, device=device) / 255)[:, None, None]  # 3,1,1
    std_bgr = (torch.tensor(std_bgr, device=device) / 255).unsqueeze(-1).unsqueeze(-1)
    img_ts = (img_ts * std_bgr + mean_bgr) * 255
    return img_ts


def f_recover_normalization4np(img_np_bgr, mean_bgr=(128.0, 128.0, 128.0),
                               std_bgr=(256.0, 256.0, 256.0)):
    '''

    :param img_np_bgr: c,h,w
    :return:
    '''
    mean_bgr = np.array(mean_bgr)
    std_bgr = np.array(std_bgr)
    img_np_bgr = img_np_bgr * std_bgr + mean_bgr
    return img_np_bgr


def x_random_rotation4ts(img_np, boxes_ltrb_ts, kps_3d_ts, degree=10):
    '''
    kps_ts: torch.Size([1, 5, 3])
    '''
    angle = random.uniform(-degree, degree)
    h, w, c = img_np.shape
    cx, cy = w / 2.0, h / 2.0

    ''' 图片处理 '''
    # mat rotate 1 center 2 angle 3 缩放系数
    matRotate = cv2.getRotationMatrix2D((cy, cx), angle, 1.0)
    img_np = cv2.warpAffine(img_np, matRotate, (h, w))

    # debug
    # print(angle)
    # plt.imshow(img_np)
    # plt.show()

    ''' box处理 '''
    a = -angle / 180.0 * math.pi
    # boxes = torch.from_numpy(boxes)
    new_boxes_ltrb_ts = torch.zeros_like(boxes_ltrb_ts)
    new_boxes_ltrb_ts[:, 0] = boxes_ltrb_ts[:, 1]
    new_boxes_ltrb_ts[:, 1] = boxes_ltrb_ts[:, 0]
    new_boxes_ltrb_ts[:, 2] = boxes_ltrb_ts[:, 3]
    new_boxes_ltrb_ts[:, 3] = boxes_ltrb_ts[:, 2]
    for i in range(boxes_ltrb_ts.shape[0]):
        ymin, xmin, ymax, xmax = new_boxes_ltrb_ts[i, :]
        xmin, ymin, xmax, ymax = float(xmin), float(ymin), float(xmax), float(ymax)
        x0, y0 = xmin, ymin
        x1, y1 = xmin, ymax
        x2, y2 = xmax, ymin
        x3, y3 = xmax, ymax
        z = torch.FloatTensor([[y0, x0], [y1, x1], [y2, x2], [y3, x3]])
        tp = torch.zeros_like(z)
        tp[:, 1] = (z[:, 1] - cx) * math.cos(a) - (z[:, 0] - cy) * math.sin(a) + cx
        tp[:, 0] = (z[:, 1] - cx) * math.sin(a) + (z[:, 0] - cy) * math.cos(a) + cy
        ymax, xmax = torch.max(tp, dim=0)[0]
        ymin, xmin = torch.min(tp, dim=0)[0]
        new_boxes_ltrb_ts[i] = torch.stack([ymin, xmin, ymax, xmax])
    new_boxes_ltrb_ts[:, 1::2].clamp_(min=0, max=h - 1)
    new_boxes_ltrb_ts[:, 0::2].clamp_(min=0, max=w - 1)
    boxes_ltrb_ts[:, 0] = new_boxes_ltrb_ts[:, 1]
    boxes_ltrb_ts[:, 1] = new_boxes_ltrb_ts[:, 0]
    boxes_ltrb_ts[:, 2] = new_boxes_ltrb_ts[:, 3]
    boxes_ltrb_ts[:, 3] = new_boxes_ltrb_ts[:, 2]
    # boxes_ts = boxes_ts.numpy()

    ngt, nkey, c = kps_3d_ts.shape
    for i in range(ngt):
        for j in range(nkey):
            x = kps_3d_ts[i][j][0]
            y = kps_3d_ts[i][j][1]
            p = np.array([x, y, 1])
            p = matRotate.dot(p)
            kps_3d_ts[i][j][0] = p[0]
            kps_3d_ts[i][j][1] = p[1]

    return img_np, boxes_ltrb_ts, kps_3d_ts


def x_random_rotation4np(img_np, boxes_ltrb_np, kps_3d_np, degree=10):
    angle = random.uniform(-degree, degree)
    h, w, c = img_np.shape
    cx, cy = w / 2.0, h / 2.0

    ''' 图片处理 '''
    # mat rotate 1 center 2 angle 3 缩放系数
    matRotate = cv2.getRotationMatrix2D((cy, cx), angle, 1.0)
    img_np = cv2.warpAffine(img_np, matRotate, (h, w))

    # debug
    # print(angle)
    # plt.imshow(img_np)
    # plt.show()

    ''' box处理 '''
    a = -angle / 180.0 * math.pi
    # boxes = torch.from_numpy(boxes)
    new_boxes_ltrb_np = np.zeros_like(boxes_ltrb_np)
    new_boxes_ltrb_np[:, 0] = boxes_ltrb_np[:, 1]
    new_boxes_ltrb_np[:, 1] = boxes_ltrb_np[:, 0]
    new_boxes_ltrb_np[:, 2] = boxes_ltrb_np[:, 3]
    new_boxes_ltrb_np[:, 3] = boxes_ltrb_np[:, 2]
    for i in range(boxes_ltrb_np.shape[0]):
        ymin, xmin, ymax, xmax = new_boxes_ltrb_np[i, :]
        xmin, ymin, xmax, ymax = float(xmin), float(ymin), float(xmax), float(ymax)
        x0, y0 = xmin, ymin
        x1, y1 = xmin, ymax
        x2, y2 = xmax, ymin
        x3, y3 = xmax, ymax
        z = np.array([[y0, x0], [y1, x1], [y2, x2], [y3, x3]], dtype=np.float32)
        tp = np.zeros_like(z)
        tp[:, 1] = (z[:, 1] - cx) * math.cos(a) - (z[:, 0] - cy) * math.sin(a) + cx
        tp[:, 0] = (z[:, 1] - cx) * math.sin(a) + (z[:, 0] - cy) * math.cos(a) + cy
        ymax, xmax = np.max(tp, axis=0)
        ymin, xmin = np.min(tp, axis=0)
        new_boxes_ltrb_np[i] = np.stack([ymin, xmin, ymax, xmax])

    new_boxes_ltrb_np[:, 1::2] = np.clip(new_boxes_ltrb_np[:, 0::2], a_min=0, a_max=h - 1)
    new_boxes_ltrb_np[:, 1::2] = np.clip(new_boxes_ltrb_np[:, 0::2], a_min=0, a_max=w - 1)
    boxes_ltrb_np[:, 0] = new_boxes_ltrb_np[:, 1]
    boxes_ltrb_np[:, 1] = new_boxes_ltrb_np[:, 0]
    boxes_ltrb_np[:, 2] = new_boxes_ltrb_np[:, 3]
    boxes_ltrb_np[:, 3] = new_boxes_ltrb_np[:, 2]

    ngt, nkey, c = kps_3d_np.shape
    for i in range(ngt):
        for j in range(nkey):
            x = kps_3d_np[i][j][0]
            y = kps_3d_np[i][j][1]
            p = np.array([x, y, 1])
            p = matRotate.dot(p)
            kps_3d_np[i][j][0] = p[0]
            kps_3d_np[i][j][1] = p[1]

    return img_np, boxes_ltrb_np, kps_3d_np


def t_transformer():
    from ftools.ai.datas.z_dataloader import get_data_voc2007, get_data_widerface, get_data_face5, \
        get_data_face98, get_data_lsp
    mode = 'keypoints'  # bbox segm keypoints caption

    # mode = 'bbox'  # bbox segm keypoints caption

    class CFG:
        pass

    path_host = 'M:'
    cfg = CFG()
    cfg.IS_TRAIN = True
    cfg.IS_VAL = False
    cfg.IS_TEST = True
    cfg.IS_MULTI_SCALE = False
    cfg.SIZE_WH_INPUT_TRAIN = (320, 320)
    cfg.IS_MOSAIC = False
    cfg.IS_MIXUP = False
    cfg.IS_MOSAIC_FILTER = False
    cfg.THR_WH = 5
    cfg.THR_AREA = 25
    cfg.NUM_WORKERS = 0

    transformer_train = FCompose(
        [
            # FRandomRotate(40),
            FRandomSampleCrop(),
            # FExpand(),
            # FResizeKeep(cfg.SIZE_WH_INPUT_TRAIN),
            # FResize(cfg.SIZE_WH_INPUT_TRAIN),
            # FRandomMirror(KEYPOINTS_Mirror['face5']),
            # Normalize_v2(),
            # to_tensor_target2one(is_box_oned=False),
        ])
    # get_data_widerface 数据中自带校验方法
    data_info, dataloader_train, dataloader_test = get_data_lsp(
        cfg, path_host=path_host,
        transform_train=transformer_train,
        mode_train=mode, batch_train=10)

    dataset = dataloader_train.dataset
    ids_classes = dataset.ids_classes

    # image, target = dataset[17]
    # f_show_kp_np4cv(image,
    #                 kps=target['kps'],
    #                 kps_seq=None,
    #                 g_ltrb=target['boxes'],
    #                 is_color_same=True,
    #                 )

    for i in range(0, 16):
        image, target = dataset[i]
        f_show_kp_np4cv(image,
                        g_kps_l=target['kps'],
                        kps_seq=None,
                        g_ltrb=target['boxes'],
                        is_color_same=True,
                        )


if __name__ == '__main__':
    class cfg:
        pass


    t_transformer()  # 测试 transformer

    # from nanodet.detect_pic import get_image_list
    #
    # path_pic = r'M:\AI\datas\coco2017\imgs\val2017_5000'
    # if os.path.isdir(path_pic):
    #     files_pic = get_image_list(path_pic)
    # else:
    #     files_pic = [path_pic]
    #
    # size_wh_input = (320, 320)
    # _transform_test = f_get_transform_val_base(size_wh_input)
    #
    # for file in files_pic:
    #     img_np_file = cv2.imread(file)  # h,w,c bgr
    #     h, w = img_np_file.shape[:2]
    #     size_wh_file = np.array([w, h])
    #
    #     img_ts_3d, target = _transform_test(img_np_file)
    #     f_show_od_ts4plt_v3(img_ts_3d, target['boxes'], is_normal=True, is_torgb=True)



import os
import random

import cv2
import torch
from pycocotools.coco import COCO
from torch.utils.data import Dataset
import numpy as np

from ftools.GLOBAL_LOG import flog
from ftools.ai.datas.dta_heighten.fmosaic import get_mosaic_img, mixup
from ftools.ai.object_detection.boxes.f_kps import fsplit_kps


class CustomCocoDataset(Dataset):
    '''

    '''

    def __init__(self, file_json, path_img, mode, transform=None, is_debug=False,
                 s_ids_cats=None, nums_cat=None, is_ts_all=True, is_jump=False,
                 size_hw_img=None, mode_balance_data=None, name='fname',
                 kps_seq=None,
                 ):
        '''

        :param mode:  bbox segm keypoints caption
        :param s_ids_cats:  指定类别 或数据集的类别ID
        :param nums_cat:  限制类别的最大数量
        :param is_ts_all:  默认全部转TS
        :param is_jump:  是否跳过没有目标的图片 True 为必须重新加载
        :param size_hw_img:
        :param mode_balance_data: 平衡各类数据量 这个优先冲突于 s_ids_cats  数据类别平衡方法
            'max': 4舍五入倍数整体复制  确保类别尽量一致
            'min': 取最少的类型数,多的随机选
            None: 不处理
            int 指定数据数量
        :param name: dataset 标注
        '''
        self.path_img = path_img
        if not os.path.exists(path_img):
            raise Exception('coco path_img 路径不存在', path_img)

        self.size_hw_img = size_hw_img
        self.is_jump = is_jump
        self.kps_seq = kps_seq

        self.file_json = file_json
        self.transform = transform
        self.mode = mode
        self.coco_obj = COCO(file_json)
        # self.device = device  # 不能扩 use CUDA with multiprocessing

        self.name = name
        self.is_ts_all = is_ts_all

        # f_look_coco_type(self.coco_obj, ids_cats_ustom=None)
        print('创建dataset-----', name)

        self.s_ids_cats = []  # 初始化一下
        # 平衡各类数据量  模式
        if mode_balance_data is not None:
            cats = self.coco_obj.getCatIds()
            num_class = len(cats)
            _t_ids = []
            _t_len = np.zeros(num_class)
            for i, cat in enumerate(cats):
                ids = self.coco_obj.getImgIds(catIds=cat)
                _t_ids.append(ids)
                _t_len[i] = len(ids)

            self.ids_img = []

            if mode_balance_data == 'max':
                num_repeats = np.around((_t_len.max() / _t_len)).astype(np.int)
                for i in range(num_class):
                    self.ids_img.extend(np.tile(np.array(_t_ids[i]), num_repeats[i]).tolist())
            elif mode_balance_data == 'min':
                len_min = _t_len.min().astype(np.int)
                # flog.debug('_t_len = %s', _t_len)
                for i in range(num_class):
                    self.ids_img.extend(np.random.choice(_t_ids[i], len_min).tolist())
            elif isinstance(mode_balance_data, int):
                for cat_id in cats:
                    # 每一类的ID
                    ids_ = self.coco_obj.getImgIds(catIds=[cat_id])
                    if len(ids_) > 0:
                        self.s_ids_cats.append(cat_id)
                    self.ids_img.extend(np.random.choice(ids_, mode_balance_data).tolist())

        else:
            # 与 mode_balance_data 冲突
            if s_ids_cats is not None:
                flog.warning('指定coco类型 %s', self.coco_obj.loadCats(s_ids_cats))
                self.s_ids_cats = s_ids_cats
                ids_img = []

                # 限制每类的最大个数
                if nums_cat is None:
                    for idc in zip(s_ids_cats):
                        # 类型对应哪些文件 可能是一张图片多个类型
                        ids_ = self.coco_obj.getImgIds(catIds=idc)
                        ids_img += ids_
                else:
                    # 限制每类的最大个数
                    for idc, num_cat in zip(s_ids_cats, nums_cat):
                        # 类型对应哪些文件 可能是一张图片多个类型
                        ids_ = self.coco_obj.getImgIds(catIds=idc)[:num_cat]
                        # ids_ = self.coco.getImgIds(catIds=idc)[:1000]
                        ids_img += ids_
                        # print(ids_)  # 这个只支持单个元素

                self.ids_img = list(set(ids_img))  # 去重
            else:
                # 所有类别所有图片
                self.s_ids_cats = self.coco_obj.getCatIds()
                self.ids_img = self.coco_obj.getImgIds()  # 所有图片的id 画图数量

        #  创建 coco 类别映射
        self._init_load_classes(self.s_ids_cats)  # 除了coco数据集,其它不管

        self.is_debug = is_debug  # 备用 预留

    def __len__(self):
        if self.is_debug:
            return 9  # 用一云debug
        return len(self.ids_img)

    def open_img_tar(self, id_img):
        img = self.load_image(id_img)

        # bboxs, labels, or keypoints  np
        tars_list = self.load_anns(id_img, img_wh=img.shape[:2][::-1])

        # 没有标注就只有这两个属性
        target = {}
        target['image_id'] = id_img
        target['size'] = np.array(img.shape[:2][::-1])  # (w,h)

        if tars_list is None:  # 没有标注返回空
            # target['boxes'] = np.empty((0, 4), dtype=np.float32)
            # target['labels'] = np.empty((0), dtype=np.float32)
            return img, target

        # 根据标注模式 及字段自动添加 target['boxes', 'labels', 'kps']
        key_names = ['boxes', 'labels', 'kps', ]
        for i, tar in enumerate(tars_list):
            target[key_names[i]] = tar
        return img, target

    def fgetitem(self, id_img):
        '''
        这个方法不要返回空
        :param id_img:
        :return:
        '''
        # 这里生成的是原图尺寸的 target 和img_np_uint8 (375, 500, 3)
        img, target = self.open_img_tar(id_img)

        if 'boxes' in target:
            assert len(target['boxes']) == len(target['labels']), \
                '!!! 数据有问题 %s  %s %s %s ' % ('transform前', target, len(target['boxes']), len(target['labels']))
        else:
            if self.is_jump:
                # print('这个图片没有标注信息 id为 %s ,跳过', id_img)
                return self.fgetitem(self.ids_img[random.randint(0, len(self.ids_img) - 1)])
        # 二级检测无用
        # if target['boxes'].shape[0] == 0:
        #     flog.warning('数据有问题 重新加载 %s', id_img)
        #     return self.fgetitem(self.ids_img[random.randint(0, len(self.ids_img) - 1)])

        return img, target

    def __getitem__(self, index):
        '''

        :param index:
        :return: tensor or np.array 根据 out: 默认ts or other is np
            img: h,w,c
            target:
            coco原装是 ltwh
            dict{
                image_id: int,
                bboxs: ts n4 原图 ltwh -> ltrb
                labels: ts n,
                keypoints: 待定
                size: wh
            }
        '''
        id_img = self.ids_img[index]
        img, target = self.fgetitem(id_img)

        # debug
        # f_show_od_ts4plt_v3(
        #     img,
        #     target['boxes'],
        #     is_normal=True,
        # )

        # 以上img 确定是np格式(transform后出来一般是ts); target 全部是np 除imgid
        if self.transform is not None:
            img, target = self.transform(img, target)

        # 这里img输出 ts_3d
        if self.is_ts_all:
            if 'boxes' in target:
                target['boxes'] = torch.tensor(target['boxes'], dtype=torch.float)
                target['labels'] = torch.tensor(target['labels'], dtype=torch.int64)
            target['size'] = torch.tensor(target['size'], dtype=torch.float)  # file尺寸

        # 每个图片对应的target数量是不一致的 所以需要用target封装
        return img, target

    def load_image(self, id_img):
        '''

        :param id_img:
        :return:
        '''
        image_info = self.coco_obj.loadImgs(id_img)[0]
        file_img = os.path.join(self.path_img, image_info['file_name'])
        if not os.path.exists(file_img):
            raise Exception('file_img 加载图片路径错误', file_img)

        img = cv2.imread(file_img)
        return img

    def load_anns(self, id_img, img_wh):
        '''
        ltwh --> ltrb
        :param id_img:
        :return: tars_list 没有或有问题返回空
            bboxs: np(num_anns, 4)
            labels: np(num_anns)
        '''
        # annotation_ids = self.coco.getAnnIds(self.image_ids[index], iscrowd=False)
        annotation_ids = self.coco_obj.getAnnIds(id_img)  # ann的id
        # anns is num_anns x 4, (x1, x2, y1, y2)

        if len(annotation_ids) == 0:
            flog.error('这图标注ID不存在 id_img = %s', id_img)
            return None

        coco_anns = self.coco_obj.loadAnns(annotation_ids)
        bboxs_np = np.zeros((0, 4), dtype=np.float32)  # np创建 空数组 高级
        labels = []
        kps = []
        ''' 默认 bbox 一定是存在的 '''
        for ann in coco_anns:
            x, y, box_w, box_h = ann['bbox']  # ltwh
            # 修正超范围的框  得 ltrb
            x1 = max(0, x)  # 修正lt最小为0 左上必须在图中
            y1 = max(0, y)
            x2 = min(img_wh[0] - 1, x1 + max(0, box_w - 1))  # 右下必须在图中
            y2 = min(img_wh[1] - 1, y1 + max(0, box_h - 1))
            ''' --- bbox校验 --- '''
            if ann['area'] > 0 and x2 > x1 and y2 >= y1:
                bbox = np.array([[x1, y1, x2, y2]], dtype=np.float32)
                # bbox[0, :4] = ann['bbox']
                # ann['bbox'] = [x1, y1, x2, y2]  # 这样写回有BUG 共享内存会修改
            else:
                flog.error('标记框有问题 %s 跳过', ann)
                continue

            if self.mode == 'keypoints':
                ''' 这里只取了 v=2的转换为1(有效用于训练)  v=0:没有  v=1:标了不可见 v=2标了可见 '''
                _kp = np.array(ann['keypoints'])
                # 校验_kps 这里只会有一个标注
                # (ngt,dim_kps) ->  (ngt,num_keypoints,3)
                _kp = _kp.reshape(-1, 3)
                mask_pos = _kp[:, 2] == 2
                _kp[:, 2] = 0  # 全部为0后 只取可见 v=0 和v=1 都改为0
                _kp[:, 2][mask_pos] = 1
                kps.append(_kp.reshape(-1))

            # 全部通过后添加
            bboxs_np = np.append(bboxs_np, bbox, axis=0)
            labels.append(self.classes_coco2train[ann['category_id']])

        if len(labels) == 0:
            return None

        # bboxs = ltwh2ltrb(bboxs) # 前面 已转
        if bboxs_np.shape[0] == 0:
            flog.error('这图标注 不存在 %s', coco_anns)
            return None
            # raise Exception('这图标注 不存在 %s', coco_anns)

        # 转NP
        labels = np.array(labels, dtype=np.float32)

        if self.mode == 'bbox':
            return bboxs_np, labels
        elif self.mode == 'keypoints':
            # 转NP
            kps = np.array(kps, dtype=np.float32)
            return bboxs_np, labels, kps

    def _init_load_classes(self, ids_cat):
        '''
        self.classes_ids :  {'Parade': 1}
        self.ids_classes :  {1: 'Parade'}
        self.ids_new_old {1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6, 7: 7, 8: 8, 9: 9, 10: 10, 11: 11, 12: 12, 13: 13, 14: 14, 15: 15, 16: 16, 17: 17, 18: 18, 19: 19, 20: 20}
        self.ids_old_new {1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6, 7: 7, 8: 8, 9: 9, 10: 10, 11: 11, 12: 12, 13: 13, 14: 14, 15: 15, 16: 16, 17: 17, 18: 18, 19: 19, 20: 20}
        :return:
        '''
        # [{'id': 1, 'name': 'aeroplane'}, {'id': 2, 'name': 'bicycle'}, {'id': 3, 'name': 'bird'}, {'id': 4, 'name': 'boat'}, {'id': 5, 'name': 'bottle'}, {'id': 6, 'name': 'bus'}, {'id': 7, 'name': 'car'}, {'id': 8, 'name': 'cat'}, {'id': 9, 'name': 'chair'}, {'id': 10, 'name': 'cow'}, {'id': 11, 'name': 'diningtable'}, {'id': 12, 'name': 'dog'}, {'id': 13, 'name': 'horse'}, {'id': 14, 'name': 'motorbike'}, {'id': 15, 'name': 'person'}, {'id': 16, 'name': 'pottedplant'}, {'id': 17, 'name': 'sheep'}, {'id': 18, 'name': 'sofa'}, {'id': 19, 'name': 'train'}, {'id': 20, 'name': 'tvmonitor'}]
        categories = self.coco_obj.loadCats(ids_cat)
        categories.sort(key=lambda x: x['id'])  # 按id升序 [{'id': 1, 'name': 'Parade'},]

        # coco ids is not from 1, and not continue ,make a new index from 0 to 79, continuely
        # 重建index 从1-80
        # classes_ids:   {names:      new_index}
        # coco_ids:  {new_index:  coco_index}
        # coco_ids_inverse: {coco_index: new_index}

        self.classes_ids, self.classes_train2coco, self.classes_coco2train = {}, {}, {}
        self.ids_classes = {}
        # 解决中间有断格的情况
        for i, c in enumerate(categories, start=1):  # 修正从1开始
            self.classes_train2coco[i] = c['id']  # 验证时用这个
            self.classes_coco2train[c['id']] = i
            self.classes_ids[c['name']] = i  # 这个是 train 索引 {'Parade': 0,}
            self.ids_classes[i] = c['name']  # 这个是 train 索引 {0:'Parade',}
        pass


class HeightenCocoDataset(CustomCocoDataset):

    def __init__(self, file_json, path_img, mode, transform=None, is_debug=False, s_ids_cats=None, nums_cat=None,
                 is_ts_all=True, is_jump=False, size_hw_img=None, mode_balance_data=None, name='fname',
                 is_mosaic=False, is_mixup=False,
                 is_mosaic_filter=False, thr_wh=5, thr_area=25,
                 kps_seq=None,
                 ):
        super().__init__(file_json=file_json, path_img=path_img, mode=mode,
                         transform=transform, is_debug=is_debug,
                         s_ids_cats=s_ids_cats, nums_cat=nums_cat,
                         is_ts_all=is_ts_all, is_jump=is_jump,
                         size_hw_img=size_hw_img,
                         mode_balance_data=mode_balance_data, name=name,
                         kps_seq=kps_seq,
                         )
        # 这个用于 mosaic 的尺寸
        assert size_hw_img is not None, 'size_hw_img 不能为空'
        self.size_hw_img = size_hw_img
        self.is_mosaic = is_mosaic
        self.is_mixup = is_mixup

        # mosaic 过滤参数 如果 transform 不带过滤 则建议开启
        self.is_mosaic_filter = is_mosaic_filter
        self.thr_wh = thr_wh
        self.thr_area = thr_area

    def fgetitem(self, id_img):
        if self.is_mosaic:
            # 这里自动强制将 np.int 转成 int coco不报错
            ids_img = [id_img] + np.random.choice(self.ids_img, 3).tolist()
            imgs, targets = [], []
            for id_img in ids_img:
                img_np, target = super().fgetitem(id_img)
                imgs.append(img_np)
                targets.append(target)

            img_np, target = get_mosaic_img(imgs=imgs,
                                            targets=targets,
                                            size_hw_m=self.size_hw_img,
                                            is_filter=self.is_mosaic_filter,
                                            thr_wh=self.thr_wh,
                                            thr_area=self.thr_area,
                                            )
            # debug
            # from ftools.ai.picture.f_show import f_show_od_np4plt_v3
            # f_show_od_np4plt_v3(img_np, g_ltrb=target['boxes'],
            #                                 g_texts=target['labels'].astype(np.int).tolist())

        elif self.is_mixup:
            img_src, target_src = super().fgetitem(id_img)
            # _id_img = np.random.choice(self.ids_img, 1)[0]  # 这个返回的是np.int
            _id_img = random.randint(0, len(self.ids_img) - 1)
            img_det, target_det = super().fgetitem(_id_img)  # dataset 已确保有目标

            img_np, target = mixup(img_src=img_src, target_src=target_src,
                                   img_det=img_det, target_det=target_det,
                                   size_hw_m=self.size_hw_img,
                                   mixup_scale=(0.5, 1.5),
                                   )
            # debug
            # from ftools.ai.picture.f_show import f_show_od_np4plt_v3
            # f_show_od_np4plt_v3(img_np, g_ltrb=target['boxes'],
            #                     g_texts=target['labels'].astype(np.int).tolist())
        else:
            img_np, target = super().fgetitem(id_img)

        return img_np, target


if __name__ == '__main__':
    from ftools.ai.datas.data_factory import CLS4collate_fn
    from ftools.ai.datas.z_dataloader import get_data_cocomin, get_data_type3
    from ftools.ai.picture.f_show import f_show_od_np4plt_v3, f_show_od_np4cv
    import numpy as np


    class CFG:
        pass


    path_host = 'M:'
    mode = 'bbox'
    cfg = CFG()
    cfg.IS_TRAIN = True
    cfg.IS_VAL = False
    cfg.IS_TEST = False
    cfg.IS_MULTI_SCALE = False
    cfg.num_workers = 0

    # mosaic 必须传
    cfg.SIZE_WH_INPUT_TRAIN = (640, 640)

    cfg.IS_MOSAIC = True
    cfg.IS_MIXUP = True
    cfg.IS_MOSAIC_FILTER = False
    cfg.THR_WH = 5
    cfg.THR_AREA = 25

    data_info, dataloader_train, dataloader_test = get_data_type3(
        cfg,
        path_host=path_host,
        mode_train=mode, batch_train=10
    )

    dataset = dataloader_train.dataset

    for data in dataset:
        img_np, target = data
        text = [str(l) for l in target['labels']]
        # f_show_od_np4plt_v3(img_np, target['boxes'], g_texts=text, )
        f_show_od_np4cv(img_np, target['boxes'], g_texts=text, )



import math
import random

import matplotlib
import numpy as np
import cv2
import torch
from PIL import Image, ImageDraw, ImageFont
from torchvision import transforms
import matplotlib.pyplot as plt

from ftools.GLOBAL_LOG import flog
from ftools.general.FColor import fcre_color_indexes, COLORS_MAP, fget_color_random, fget_color_val, COLORS_RGB, \
    fhex2rgb

'''
plt 常用颜色: 'lightgreen' 'red' 'tan'
'''

FONT4CV = cv2.FONT_HERSHEY_SIMPLEX  # 加粗
FONT4CV = cv2.FONT_HERSHEY_COMPLEX  # 细点


def _draw_grid4plt(size, grids):
    '''

    :param size:
    :param grids:
    :return:
    '''
    # colors_ = STANDARD_COLORS[random.randint(0, len(STANDARD_COLORS) - 1)]
    colors_ = 'Pink'
    w, h = size
    xys = np.array([w, h]) / grids
    off_x = np.arange(1, grids[0])
    off_y = np.arange(1, grids[1])
    xx = off_x * xys[0]
    yy = off_y * xys[1]

    for x_ in xx:
        # 画列
        plt.plot([x_, x_], [0, h], color=colors_, linewidth=1., alpha=0.3)
    for y_ in yy:
        # 画列
        plt.plot([0, w], [y_, y_], color=colors_, linewidth=1., alpha=0.3)


def _convert_uint8(img_np):
    if img_np.dtype is not np.uint8:
        img_np_uint8 = img_np.copy()
        img_np_uint8 = img_np_uint8.astype(np.uint8)
    else:
        return img_np
    return img_np_uint8


def _draw_box4plt(boxes, texts=None, font_size=10, color='red', ax=None, recover_sizewh=None, linewidth=1):
    '''

    :param boxes:
    :param texts: 与 boxes 对应的 文本 list
    :param font_size:
    :param color:
    :param ax:  ax = plt.gca()
    :return:
    '''
    if recover_sizewh is not None:
        whwh = np.tile(np.array(recover_sizewh), 2)  # 整体复制 tile
        boxes = boxes * whwh

    if texts is not None:
        try:
            font = ImageFont.truetype('simhei.ttf', font_size, encoding='utf-8')  # 参数1：字体文件路径，参数2：字体大小
        except IOError:
            font = ImageFont.load_default()
        text_w, text_h = font.getsize(str(texts[0]))
        margin = np.ceil(0.05 * text_h)

    for i, box in enumerate(boxes):
        l, t, r, b = box
        w = r - l
        h = b - t

        if ax is None:
            plt.Rectangle((l, t), w, h, color=color, fill=False, linewidth=linewidth)
        else:
            ax.add_patch(plt.Rectangle((l, t), w, h, color=color, fill=False, linewidth=linewidth))

        if texts is not None:
            if ax is None:
                plt.text(l + margin, t - text_h - margin, str(texts[i]),
                         color="r", ha='center', va='center',
                         bbox={'facecolor': 'blue', 'alpha': 0.5, 'pad': 1})
            else:
                ax.text(l + margin, t - text_h - margin, str(texts[i]),
                        color="r", ha='center', va='center',
                        bbox={'facecolor': 'blue', 'alpha': 0.5, 'pad': 1})

        x = l + w / 2
        y = t + h / 2
        plt.scatter(x, y, marker='x', color=color, s=40, label='First')


def f_show_od_ts4plt_v3(img_ts, g_ltrb=None, p_ltrb=None, is_recover_size=False,
                        p_texts=None, g_texts=None, is_normal=False, is_torgb=False,
                        grid_wh_np=None, title=None):
    '''

    :param img_ts:
    :param g_ltrb:
    :param p_ltrb:
    :param is_recover_size:
    :param p_texts:  输入 list
    :param g_texts:
    :param is_normal:
    :param is_torgb:
    :param grid_wh_np: np(7,7)
    :param title: 指定标题 不指定自动生成
    :return:
    '''
    assert isinstance(img_ts, torch.Tensor)
    # c,h,w -> h,w,c
    if is_normal:
        from ftools.ai.datas.dta_heighten.f_data_heighten4target import f_recover_normalization4ts_v2
        img_ts = img_ts.clone()
        img_ts = f_recover_normalization4ts_v2(img_ts)
    img_np = img_ts.cpu().numpy().astype(np.uint8).transpose((1, 2, 0))
    if is_torgb:
        img_np = cv2.cvtColor(img_np, cv2.COLOR_BGR2RGB)
    f_show_od_np4plt_v3(img_np, g_ltrb=g_ltrb, p_ltrb=p_ltrb,
                        is_recover_size=is_recover_size,
                        title=title,
                        p_texts=p_texts, g_texts=g_texts, grid_wh_np=grid_wh_np)


def f_show_od_np4plt_v3(img_np, g_ltrb=None, p_ltrb=None, other_ltrb=None, title=None,
                        is_recover_size=False, p_texts=None, g_texts=None, grid_wh_np=None,
                        ):
    img_np = _convert_uint8(img_np)
    plt.imshow(img_np)
    # plt.show()
    ax = plt.gca()

    if is_recover_size:
        recover_sizewh = img_np.shape[:2][::-1]  # npwh
    else:
        recover_sizewh = None

    if grid_wh_np is not None:
        wh = img_np.shape[:2][::-1]  # npwh
        _draw_grid4plt(wh, grid_wh_np)

    _draw_title4plt(img_np, p_ltrb, title)

    if other_ltrb is not None:
        _draw_box4plt(other_ltrb, color='tan', ax=ax, recover_sizewh=recover_sizewh)

    if g_ltrb is not None:
        _draw_box4plt(g_ltrb, texts=g_texts, color='lightgreen', ax=ax, recover_sizewh=recover_sizewh, linewidth=3)

    if p_ltrb is not None:
        _draw_box4plt(p_ltrb, texts=p_texts, color='red', ax=ax, recover_sizewh=recover_sizewh)

    plt.show()


def _draw_title4plt(img_np, p_ltrb, title):
    if title is not None:
        text = title
    else:
        if p_ltrb is not None:
            text = '%s x %s (num_pos = %s)' % (str(img_np.shape[1]), str(img_np.shape[0]), str(len(p_ltrb)))
        else:
            text = '%s x %s ' % (str(img_np.shape[1]), str(img_np.shape[0]))
    plt.title(text)


def f_show_kp_np4plt(img_np, kps, kps_seq=None, g_ltrb=None, p_ltrb=None, other_ltrb=None, title=None,
                     is_recover_size=False, p_texts=None, g_texts=None, kps_size=20,
                     ):
    indexes_color = fcre_color_indexes(len(kps[0]) // 3)

    img_np = _convert_uint8(img_np)
    plt.imshow(img_np)
    # plt.show()
    ax = plt.gca()

    if is_recover_size:
        recover_sizewh = img_np.shape[:2][::-1]  # npwh
    else:
        recover_sizewh = None

    _draw_title4plt(img_np, p_ltrb, title)
    for kp in kps:  # 这是一批
        for i, open in enumerate(kp[2::3]):
            if open != 0:
                plt.scatter(kp[::3], kp[1::3],
                            color=np.array(list(COLORS_MAP.keys()))[indexes_color],
                            s=kps_size, alpha=1.)
        if kps_seq is not None:
            for seq in kps_seq:
                _kp = kp.reshape(-1, 3)
                xyt_s = _kp[seq[0]]
                xyt_e = _kp[seq[1]]
                if xyt_s[2] != 0 and xyt_e[2] != 0:
                    plt.plot([xyt_s[0], xyt_e[0]],
                             [xyt_s[1], xyt_e[1]],
                             color=list(COLORS_MAP.keys())[indexes_color[seq[0]]],
                             linewidth=2,
                             alpha=0.8,
                             )

    if other_ltrb is not None:
        _draw_box4plt(other_ltrb, color='tan', ax=ax, recover_sizewh=recover_sizewh)

    if g_ltrb is not None:
        _draw_box4plt(g_ltrb, texts=g_texts, color='lightgreen', ax=ax, recover_sizewh=recover_sizewh, linewidth=3)

    if p_ltrb is not None:
        _draw_box4plt(p_ltrb, texts=p_texts, color='red', ax=ax, recover_sizewh=recover_sizewh)


def draw_text_chinese4cv(img_np, text, left, top, textColor=(0, 255, 0), textSize=20):
    # 图片，添加的文字，左上角坐标，字体，字体大小，颜色，字体粗细
    # img_np = cv2.putText(img_np, text_, (0, 40), cv2.FONT_HERSHEY_SIMPLEX, 0.5,
    #                      (0, 255, 0), 1)
    img_pil = Image.fromarray(cv2.cvtColor(img_np, cv2.COLOR_BGR2RGB))
    # 创建一个可以在给定图像上绘图的对象
    draw = ImageDraw.Draw(img_pil)
    # 字体的格式
    fontStyle = ImageFont.truetype("font/simsun.ttc", textSize, encoding="utf-8")
    # 绘制文本
    draw.text((left, top), text, textColor, font=fontStyle)
    # 转换回OpenCV格式
    return cv2.cvtColor(np.asarray(img_pil), cv2.COLOR_RGB2BGR)
    # return img_np


def f_show_pic_np4plt(img_np):
    '''
    不支持float32
    :param pic:
    :return:
    '''
    img_np = _convert_uint8(img_np)
    plt.imshow(img_np)
    plt.show()


''' ********************   cv区 ******************** '''


def _draw_box4cv(img_np, boxes, texts=None,
                 color=fget_color_val(0), recover_sizewh=None, thickness=2):
    if recover_sizewh is not None:
        whwh = np.tile(np.array(recover_sizewh), 2)  # 整体复制 tile
        boxes = boxes * whwh

    if texts is not None:
        txt_size = cv2.getTextSize(texts[0], FONT4CV, fontScale=2, thickness=1)[0]

    for i, box in enumerate(boxes):
        box = list(map(int, box))
        l, t, r, b = box
        w = r - l
        h = b - t

        cv2.rectangle(img_np, (l, t), (r, b), color, thickness=thickness)
        if texts is not None:
            # mask_img = cv2.addWeighted(image, alpha, zeros_mask, beta, gamma) # 透明有点麻烦
            # cv2.rectangle(
            #     img_np,
            #     (l, t + 1),
            #     (l + txt_size[0] + 1, t + int(1.5 * txt_size[1])),
            #     (COLORS_RGB[3] * 255 * 0.7).astype(np.uint8).tolist()[1],
            #     thickness=-1,  # 负1填充
            # )
            cv2.putText(img_np, texts[i], (l, t + txt_size[1]), FONT4CV,
                        fontScale=1.3, color=fget_color_val(1), thickness=3)

        x = l + w / 2
        y = t + h / 2
        cv2.circle(img_np, (int(x), int(y)), radius=3,
                   color=color, thickness=-1,  # 负1填充
                   )


def _draw_title4cv(img_np, p_ltrb, title):
    h, w, _ = img_np.shape
    if title is not None:
        text = title
    else:
        if p_ltrb is not None:
            text = '%s x %s (num_pos = %s)' % (str(w), str(h), str(len(p_ltrb)))
        else:
            text = '%s x %s ' % (str(h), str(w))
    # 返回 ((宽,高),字号)
    text_scale = 0.4 * (float(w) / 300)
    txt_size = cv2.getTextSize(text, FONT4CV, text_scale, 1)[0]
    x = 0
    y = 0
    cv2.putText(img_np, text,
                (x, y + txt_size[1]), FONT4CV, text_scale,
                # fget_color_random(),
                fget_color_val(0),
                lineType=cv2.LINE_4,
                thickness=1, )


def f_show_pic_np4cv(img_np):
    img_np = _convert_uint8(img_np)
    cv2.imshow('Example', img_np)
    cv2.waitKey(0)


def f_show_kp_np4cv(img_np, p_kps_l, g_kps_l=None, kps_seq=None, g_ltrb=None, p_ltrb=None, title=None,
                    is_recover_size=False, p_texts=None, g_texts=None,
                    is_show=True, is_color_same=False,
                    ):
    '''

    :param img_np:
    :param g_kps_l:  (ngt,num_keypoints*3)
    :param kps_seq: [[13, 12], [12, 9], [12, 8], [9, 10], [8, 7], [10, 11], [7, 6], [12, 3], [12, 2], [2, 1], [1, 0], [3, 4],[4, 5]],
    :param g_ltrb:
    :param p_ltrb:
    :param title:
    :param is_recover_size:
    :param p_texts:
    :param g_texts:
    :param is_show:
    :return:
    '''
    img_np = f_show_od_np4cv(img_np, g_ltrb=g_ltrb, p_ltrb=p_ltrb, title=title,
                             is_recover_size=is_recover_size,
                             p_texts=p_texts, g_texts=g_texts,
                             is_show=False)

    num_invalid_points_g = 0
    if g_kps_l is not None:
        num_invalid_points_g = _draw_kps4cv(
            g_kps_l, img_np, is_color_same, is_recover_size, kps_seq, idx_color=1)

    num_invalid_points_p = 0
    if p_kps_l is not None:
        # 蓝色
        num_invalid_points_p = _draw_kps4cv(
            p_kps_l, img_np, is_color_same, is_recover_size, kps_seq, idx_color=5)

    if is_show:
        flog.debug('无效点个数 num_invalid_points_p={}, num_invalid_points_g={}'
                   .format(num_invalid_points_p, num_invalid_points_g))
        plt.imshow(img_np)
        plt.show()
    return img_np


def _draw_kps4cv(kps_l, img_np, is_color_same, is_recover_size, kps_seq, idx_color=5):
    '''

    :param kps_l:
    :param img_np:
    :param is_color_same:
    :param is_recover_size:
    :param kps_seq:
    :param idx_color:  相同颜色的 序列
    :return:
    '''
    num_keypoints = len(kps_l[0]) // 3

    if is_color_same:
        indexes_color = [idx_color] * num_keypoints
    else:
        indexes_color = fcre_color_indexes(num_keypoints)

    if is_recover_size:
        recover_sizewh = list(img_np.shape[:2][::-1])  # npwh
        recover_sizewh.append(1)  # 增加1维
        whwh = np.tile(np.array(recover_sizewh), len(indexes_color)).reshape(1, -1)  # 整体复制 tile
        kps_l = kps_l * whwh

    num_invalid_points = 0
    # kps_l = kps_l.astype(np.int)  # 只支持int 这里不能直接转  否则分数没有了
    for kp in kps_l:  # 这是一批
        if kps_seq is not None:
            for seq in kps_seq:
                _kp = kp.reshape(-1, 3)
                xyt_s = _kp[seq[0]]
                xyt_e = _kp[seq[1]]
                if xyt_s[2] != 0 and xyt_e[2] != 0:
                    cv2.line(img_np,
                             pt1=(int(xyt_s[0]), int(xyt_s[1])),
                             pt2=(int(xyt_e[0]), int(xyt_e[1])),
                             color=fget_color_val(indexes_color[seq[0]]),
                             thickness=1,
                             lineType=cv2.LINE_8,  # 速度：LINE_8>LINE_AA  美观：LINE_AA>LINE_8
                             )
                pass

        for i, xyt in enumerate(kp.reshape(-1, 3)):  # 这里是一个点
            if xyt[2] != 0:
                cv2.circle(img_np,
                           center=(int(xyt[0]), int(xyt[1])),
                           # center=tuple(xyt[:2].tolist()),  # 必须是tuple
                           radius=3,
                           thickness=-1,
                           color=fget_color_val(indexes_color[i])
                           )
            else:
                num_invalid_points += 1
    return num_invalid_points


def f_show_od_np4cv(img_np, g_ltrb=None, p_ltrb=None, title=None,
                    is_recover_size=False, p_texts=None, g_texts=None,
                    is_show=True,
                    ):
    img_np = _convert_uint8(img_np)

    if is_recover_size:
        recover_sizewh = img_np.shape[:2][::-1]  # npwh
    else:
        recover_sizewh = None

    _draw_title4cv(img_np, p_ltrb, title)

    if g_ltrb is not None:
        _draw_box4cv(img_np, g_ltrb, texts=g_texts, color=fhex2rgb(COLORS_MAP['lightgreen']),
                     recover_sizewh=recover_sizewh,
                     thickness=3)

    if p_ltrb is not None:
        _draw_box4cv(img_np, p_ltrb, texts=p_texts, color=fhex2rgb(COLORS_MAP['red']),
                     recover_sizewh=recover_sizewh,
                     thickness=3)

    if is_show:
        plt.imshow(img_np)
        plt.show()
    return img_np


''' ********************   cv区结束 ******************** '''


def keypoint_painter(images, maps, img_h, img_w, numpy_array=False,
                     phase_gt=False, center_map=None):
    images = images.clone().cpu().data.numpy().transpose([0, 2, 3, 1])
    maps = maps.clone().cpu().data.numpy()
    if center_map is not None:
        center_map = center_map.clone().cpu().data.numpy()
    imgs_tensor = []
    if phase_gt:
        for img, map, c_map in zip(images, maps, center_map):
            img = cv2.resize(img, (img_w, img_h))
            for m in map[:14]:
                h, w = np.unravel_index(m.argmax(), m.shape)
                x = int(w * img_w / m.shape[1])
                y = int(h * img_h / m.shape[0])
                img = cv2.circle(img.copy(), (x, y), radius=1, thickness=2, color=(255, 0, 0))
            h, w = np.unravel_index(c_map.argmax(), c_map.shape)
            x = int(w * img_w / c_map.shape[1])
            y = int(h * img_h / c_map.shape[0])
            img = cv2.circle(img.copy(), (x, y), radius=1, thickness=2, color=(0, 0, 255))
            if numpy_array:
                imgs_tensor.append(img.astype(np.uint8))
            else:
                imgs_tensor.append(transforms.ToTensor()(img))
    else:

        for img, map_6 in zip(images, maps):
            img = cv2.resize(img, (img_w, img_h))
            for step_map in map_6:
                img_copy = img.copy()
                for m in step_map[:14]:
                    h, w = np.unravel_index(m.argmax(), m.shape)
                    x = int(w * img_w / m.shape[1])
                    y = int(h * img_h / m.shape[0])
                    img_copy = cv2.circle(img_copy.copy(), (x, y), radius=1, thickness=2, color=(255, 0, 0))
                if numpy_array:
                    imgs_tensor.append(img_copy.astype(np.uint8))
                else:
                    imgs_tensor.append(transforms.ToTensor()(img_copy))
    return imgs_tensor


def f_show_cpm4input(img_np, kps_xy_input, heatmap_center_input):
    '''

    :param img_np:
    :param kps_xy_input: (14, 2)
    :param heatmap_center_input: (368, 368)
    :return:
    '''
    img_np = _convert_uint8(img_np)
    plt.imshow(img_np, alpha=0.7)
    for xy in kps_xy_input:
        plt.scatter(xy[0], xy[1], color='r', s=5, alpha=0.5)
    plt.imshow(heatmap_center_input, alpha=0.5)
    plt.show()


def f_show_cpm4t_all(img_np, kps_xy_input, heatmap_center_input, heatmap_t, size_wh_input):
    '''
    弄到input进行处理
    :param img_np:
    :param kps_xy_t: (14, 2)
    :param heatmap_center_input: (368, 368,1)
    :param heatmap_t: (46, 46, 14)
    :param size_wh_t: (46, 46, 15)
    :return:
    '''
    img_np = _convert_uint8(img_np)
    plt.imshow(img_np, alpha=1.0)
    for xy in kps_xy_input:
        plt.scatter(xy[0], xy[1], color='r', s=5, alpha=0.5)
    plt.imshow(heatmap_center_input, alpha=0.3)

    h, w, c = heatmap_t.shape
    img_heatmap_z = np.zeros((size_wh_input[1], size_wh_input[0], 1), dtype=np.uint8)
    for i in range(c):
        img_heatmap = heatmap_t[..., i][..., None]
        img_heatmap = cv2.resize(img_heatmap, size_wh_input)  # wh
        # img_heatmap = _convert_uint8(img_heatmap)
        img_heatmap_z = np.maximum(img_heatmap_z, img_heatmap[..., None])

    plt.imshow(img_heatmap_z, alpha=0.3)
    plt.show()


if __name__ == '__main__':
    file_img = r'D:\tb\tb\ai_code\fkeypoint\_test_pic\street.jpg'
    img_np_bgr = cv2.imread(file_img)
    boxes = np.array([
        [1, 1, 100, 100],
        [120, 120, 200, 200],
    ])
    gtexts = ['123', '445566']

    f_show_od_np4plt_v3(img_np_bgr, g_ltrb=boxes, g_texts=gtexts)



import cv2
import glob
import numpy as np

'''
opencv 张正友的相机标定
https://blog.csdn.net/qq_42399848/article/details/89298212
'''
if __name__ == '__main__':
    # 设置寻找亚像素角点的参数，采用的停止准则是最大循环次数30和最大误差容限0.001
    criteria = (cv2.TERM_CRITERIA_MAX_ITER | cv2.TERM_CRITERIA_EPS, 30, 0.001)
    # criteria = (cv2.TERM_CRITERIA_MAX_ITER + cv2.TERM_CRITERIA_EPS, 30, 0.001)

    # 棋盘规格
    cbraw = 9
    cbcol = 7
    path_img = 'E:/image/Ex4/*.jpg'

    # prepare object points, like (0,0,0), (1,0,0), (2,0,0) ....,(6,5,0)
    objp = np.zeros((cbraw * cbcol, 3), np.float32)
    '''
    设定世界坐标下点的坐标值，因为用的是棋盘可以直接按网格取；
    假定棋盘正好在x-y平面上，这样z=0，简化初始化步骤。
    mgrid把列向量[0:cbraw]复制了cbcol列，把行向量[0:cbcol]复制了cbraw行。
    转置reshape后，每行都是9*11网格中的某个点的坐标。
    '''
    objp[:, :2] = np.mgrid[0:cbraw, 0:cbcol].T.reshape(-1, 2)

    objpoints = []  # 存储3D点 3d point in real world space
    imgpoints = []  # 存储2D点2d points in image plane.
    # glob是个文件名管理工具
    images = glob.glob(path_img)
    for fname in images:
        # 对每张图片，识别出角点，记录世界物体坐标和图像坐标
        img = cv2.imread(fname)  # source image
        # 我用的图片太大，缩小了一半
        img = cv2.resize(img, None, fx=0.5, fy=0.5, interpolation=cv2.INTER_CUBIC)
        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)  # 转灰度
        # cv2.imshow('img',gray)
        # cv2.waitKey(1000)
        # 寻找角点，存入corners，ret是找到角点的flag  第一个参数为图片，第二个为图片横纵角点的个数
        ret, corners = cv2.findChessboardCorners(gray, (cbraw, cbcol), None)
        # 执行亚像素级角点检测 格子大小 单位mm 测量得到
        corners2 = cv2.cornerSubPix(gray, corners, (11, 11), (-1, -1), criteria)

        objpoints.append(objp)
        imgpoints.append(corners2)
        # 在棋盘上绘制角点,只是可视化工具
        img = cv2.drawChessboardCorners(gray, (cbraw, cbcol), corners2, ret)
        cv2.imshow('img', img)
        # cv2.waitKey(1000)

    cv2.destroyAllWindows()
    '''
    传入所有图片各自角点的三维、二维坐标，相机标定。
    每张图片都有自己的旋转和平移矩阵，但是相机内参和畸变系数只有一组。
    mtx，相机内参；dist，畸变系数；revcs，旋转矩阵；tvecs，平移矩阵。
    '''
    ret, mtx, dist, rvecs, tvecs = cv2.calibrateCamera(objpoints, imgpoints, gray.shape[::-1], None, None)

    print("-----------------------------------------------------")

    img = cv2.imread('E:/image/Ex4/1.jpg')
    # 注意这里跟循环开头读取图片一样，如果图片太大要同比例缩放，不然后面优化相机内参肯定是错的。
    img = cv2.resize(img, None, fx=0.5, fy=0.5, interpolation=cv2.INTER_CUBIC)
    h, w = img.shape[:2]

    '''
    优化相机内参（camera matrix），这一步可选。
    参数1表示保留所有像素点，同时可能引入黑色像素， 显示更大范围的图片（正常重映射之后会删掉一部分图像）
    设为0表示尽可能裁剪不想要的像素，这是个scale，0-1都可以取。
    '''
    newcameramtx, roi = cv2.getOptimalNewCameraMatrix(mtx, dist, (w, h), 1, (w, h))
    # 纠正畸变 方法一
    dst = cv2.undistort(img, mtx, dist, None, newcameramtx)

    # 纠正畸变 方法二  首先找到从扭曲图像到非扭曲图像的映射函数。然后使用重测函数
    # mapx, mapy = cv2.initUndistortRectifyMap(mtx, dist, None, newcameramtx, (w, h), 5)
    # dst = cv2.remap(img, mapx, mapy, cv2.INTER_LINEAR)

    # 输出纠正畸变以后的图片
    x, y, w, h = roi
    dst = dst[y:y + h, x:x + w]

    cv2.imwrite('Ex4result.png', dst)
    # 打印我们要求的两个矩阵参数
    print("newcameramtx:\n", newcameramtx)
    print("dist:\n", dist)

    # -------------------- 计算误差 ----------------------
    tot_error = 0
    for i in range(len(objpoints)):
        imgpoints2, _ = cv2.projectPoints(objpoints[i], rvecs[i], tvecs[i], mtx, dist)
        error = cv2.norm(imgpoints[i], imgpoints2, cv2.NORM_L2) / len(imgpoints2)
        tot_error += error

    print("total error mean: ", tot_error / len(objpoints))

    # 保存
    # np.savez()
    # np.savetxt()



import copy
import json
import tempfile

from pycocotools.cocoeval import COCOeval
import numpy as np

from ftools.ai.datas.z_dataloader import get_data_face5

'''
https://blog.csdn.net/u014734886/article/details/78831382
'segm', 'bbox', 'keypoints'
---------- OD  bbox --------------
[{
    "image_id": int,
    "category_id": int,
    "bbox": [x,y,width,height],
    "score": float,

}]

---------- OD  segm --------------
[{
    "image_id": int,
    "category_id": int,
    "segmentation": RLE,
    "score": float,

}]

---------- KP  keypoints --------------
[{
    "image_id": int,
    "category_id": int,
    "keypoints": [x1,y1,v1,...,xk,yk,vk],
    "score": float,
}]

---------- Stuff segmentation --------------
[{
    "image_id": int,
    "category_id": int,
    "segmentation": RLE,
}]

---------- Caption generation --------------
[{
    "image_id": int,
    "caption": str,
}]
'''


class FCOCOeval(COCOeval):
    '''
    添加了每一个类的得分
    原装生成运行后自动赋值 coco_eval_obj.stats
    '''

    def __init__(self, cocoGt=None, cocoDt=None, iouType='segm'):
        super().__init__(cocoGt, cocoDt, iouType)

    def summarize(self, catId=None):
        """
        Compute and display summary metrics for evaluation results.
        Note this functin can *only* be applied on the default parameter setting
        """

        def _summarize(ap=1, iouThr=None, areaRng='all', maxDets=100):
            p = self.params
            iStr = ' {:<18} {} @[ IoU={:<9} | area={:>6s} | maxDets={:>3d} ] = {:0.3f}'
            titleStr = 'Average Precision' if ap == 1 else 'Average Recall'
            typeStr = '(AP)' if ap == 1 else '(AR)'
            iouStr = '{:0.2f}:{:0.2f}'.format(p.iouThrs[0], p.iouThrs[-1]) \
                if iouThr is None else '{:0.2f}'.format(iouThr)

            aind = [i for i, aRng in enumerate(p.areaRngLbl) if aRng == areaRng]
            mind = [i for i, mDet in enumerate(p.maxDets) if mDet == maxDets]

            if ap == 1:
                # dimension of precision: [TxRxKxAxM]
                s = self.eval['precision']
                # IoU
                if iouThr is not None:
                    t = np.where(iouThr == p.iouThrs)[0]
                    s = s[t]

                ''' 这里是添加的代码 判断是否传入catId，如果传入就计算指定类别的指标 '''
                if isinstance(catId, int):
                    s = s[:, :, catId, aind, mind]
                else:
                    s = s[:, :, :, aind, mind]

            else:
                # dimension of recall: [TxKxAxM]
                s = self.eval['recall']
                if iouThr is not None:
                    t = np.where(iouThr == p.iouThrs)[0]
                    s = s[t]

                ''' 这里是添加的代码 判断是否传入catId，如果传入就计算指定类别的指标 '''
                if isinstance(catId, int):
                    s = s[:, catId, aind, mind]
                else:
                    s = s[:, :, aind, mind]

            if len(s[s > -1]) == 0:
                mean_s = -1
            else:
                mean_s = np.mean(s[s > -1])

            ''' 这里是修改 原始只有一个返回值 '''
            # print(iStr.format(titleStr, typeStr, iouStr, areaRng, maxDets, mean_s))
            # return mean_s
            print_string = iStr.format(titleStr, typeStr, iouStr, areaRng, maxDets, mean_s)
            return mean_s, print_string

        stats, print_list = [0] * 12, [""] * 12
        stats[0], print_list[0] = _summarize(1)
        stats[1], print_list[1] = _summarize(1, iouThr=.5, maxDets=self.params.maxDets[2])
        stats[2], print_list[2] = _summarize(1, iouThr=.75, maxDets=self.params.maxDets[2])
        stats[3], print_list[3] = _summarize(1, areaRng='small', maxDets=self.params.maxDets[2])
        stats[4], print_list[4] = _summarize(1, areaRng='medium', maxDets=self.params.maxDets[2])
        stats[5], print_list[5] = _summarize(1, areaRng='large', maxDets=self.params.maxDets[2])
        stats[6], print_list[6] = _summarize(0, maxDets=self.params.maxDets[0])
        stats[7], print_list[7] = _summarize(0, maxDets=self.params.maxDets[1])
        stats[8], print_list[8] = _summarize(0, maxDets=self.params.maxDets[2])
        stats[9], print_list[9] = _summarize(0, areaRng='small', maxDets=self.params.maxDets[2])
        stats[10], print_list[10] = _summarize(0, areaRng='medium', maxDets=self.params.maxDets[2])
        stats[11], print_list[11] = _summarize(0, areaRng='large', maxDets=self.params.maxDets[2])

        print_info = "\n".join(print_list)

        if not self.eval:
            raise Exception('Please run accumulate() first')

        # if len(self.stats)==0:
        #     self.stats = stats  # 与原装只写一次保持一致

        return stats, print_info

    def print_clses(self, clses_name):
        # calculate COCO info for all classes
        # coco_stats, print_coco = self.summarize()

        # calculate voc info for every classes(IoU=0.5)
        voc_map_info_list = []
        for i in range(len(clses_name)):
            stats, _ = self.summarize(catId=i)
            voc_map_info_list.append(" {:15}: {:.3f} {:.3f}".format(clses_name[i], stats[1], stats[7]))

        print_voc = "\n".join(voc_map_info_list)
        print(print_voc)

        # # 将验证结果保存至txt文件中
        # with open("record_mAP.txt", "w") as f:
        #     record_lines = ["COCO results:",
        #                     print_coco,
        #                     "",
        #                     "mAP(IoU=0.5) for each category:",
        #                     print_voc]
        #     f.write("\n".join(record_lines))


if __name__ == '__main__':
    annType = ['segm', 'bbox', 'keypoints']
    mode_test = annType[2]


    class CFG:
        pass


    path_host = 'M:'
    cfg = CFG()
    cfg.IS_TRAIN = True
    cfg.IS_VAL = False
    cfg.IS_TEST = True
    cfg.IS_MULTI_SCALE = False
    cfg.SIZE_WH_INPUT_TRAIN = (320, 320)
    cfg.IS_MOSAIC = False
    cfg.IS_MIXUP = False
    cfg.IS_MOSAIC_FILTER = False
    cfg.THR_WH = 5
    cfg.THR_AREA = 25
    cfg.NUM_WORKERS = 0
    cfg.IS_JUMP = True
    cfg.IS_DEBUG = False

    data_info, dataloader_train, dataloader_test = get_data_face5(
        cfg, path_host=path_host,
        mode_train=mode_test, batch_train=10)

    dataset = dataloader_train.dataset
    coco_obj_gt = dataset.coco_obj
    # imgIds = sorted(coco_obj_gt.getImgIds())
    imgIds = coco_obj_gt.getImgIds()
    # imgIds = imgIds[0:100]
    res_coco_kps = []  # 标准结果文档
    for id in imgIds:
        info_img = coco_obj_gt.loadImgs(id)[0]
        # file = os.path.join(path_img, info_img['file_name'])
        info_anns = coco_obj_gt.loadAnns(coco_obj_gt.getAnnIds(id))
        for ann in info_anns:
            if mode_test == 'bbox':
                _d = {
                    "image_id": id,
                    "category_id": ann['category_id'],
                    "bbox": ann['bbox'],
                    "score": 0.9,

                }
            elif mode_test == 'keypoints':
                _d = {
                    "image_id": id,
                    "category_id": ann['category_id'],
                    "keypoints": ann['keypoints'],  # kps_l
                    "score": 0.9,
                }
            else:
                raise Exception('暂不支持 mode_test=%s' % mode_test)

            res_coco_kps.append(_d)

    _, tmp = tempfile.mkstemp()  # 创建临时文件
    json.dump(res_coco_kps, open(tmp, 'w'))
    coco_dt = coco_obj_gt.loadRes(tmp)

    cocoEval = COCOeval(coco_obj_gt, coco_dt, mode_test)
    cocoEval.params.imgIds = imgIds

    '''
    # 标注效果 归一化因子 眼睛鼻子 .25 ~ 臀部 1.07 
    "nose",    "left_eye",    "right_eye",    "left_ear",臂    "right_ear",
    "left_shoulder", 肩   "right_shoulder",    "left_elbow",肘    "right_elbow",
    "left_wrist", 手腕   "right_wrist",    "left_hip",臀部    "right_hip",   
     "left_knee", 膝盖   "right_knee",    "left_ankle", 踝(89)   "right_ankle" 
    '''
    # cocoEval.params.kpt_oks_sigmas = np.array([.26, .25, .25, .35, .35, .
    # 79, .79, .72, .72, .62,.62,
    # 1.07, 1.07, .
    # 87, .87, .89, .89])/10.0
    cocoEval.params.kpt_oks_sigmas = np.array([.26, .25, .25, .26, .26, ]) / 10.0
    cocoEval.evaluate()
    cocoEval.accumulate()
    summarize = cocoEval.summarize()
    print(cocoEval.stats)
